import numpy as np
import onnxslim.third_party.onnx_graphsurgeon as osg
from onnxslim.core.pattern import Pattern, PatternMatcher
from onnxslim.core.pattern.registry import register_fusion_pattern


class GeluPatternMatcher(PatternMatcher):
    def __init__(self, priority):
        """Initializes a `GeluPatternMatcher` to identify and fuse GELU patterns in a computational graph."""
        pattern = Pattern(
            """
            input  input  0 2 mul_0 div_0
            Div    div_0  2 1 input ? erf_0
            Erf    erf_0  1 1 div_0 add_0
            Add    add_0  2 1 erf_0 ? mul_0
            Mul    mul_0  2 1 input add_0 mul_1
            Mul    mul_1  2 1 mul_0 ? output
            output output 1 0 mul_1
            """
        )
        super().__init__(pattern, priority)

    @property
    def name(self):
        """Returns the name of the fusion pattern, 'FusionGelu'."""
        return "FusionGelu"

    def rewrite(self, opset=11):
        """Rewrite the computation graph pattern to fuse GELU operations."""
        input_variable = self.div_0.inputs[0]
        mul_node = self.mul_0
        div_node = self.div_0

        if (
            isinstance(div_node.inputs[1], osg.Constant)
            and div_node.inputs[1].values is not None
            and np.allclose(1.4142135381698608, div_node.inputs[1].values)
        ):
            input_variable.outputs.remove(mul_node)
            input_variable.outputs.remove(div_node)

            output_variable = self.mul_1.outputs[0]
            output_variable.inputs.clear()

            return {
                self.mul_1.name: {
                    "op": "Gelu",
                    "inputs": [input_variable],
                    "outputs": [output_variable],
                    "domain": "com.microsoft",
                    "name": self.mul_1.name,
                }
            }
        else:
            return {}


register_fusion_pattern(GeluPatternMatcher(1))



#!/usr/bin/env python3
# Copyright (c) 2023 SpacemiT. All rights reserved.
import copy
import json
import os
import time
from collections import OrderedDict, deque
from typing import Dict, List, Optional, Sequence, Union

import numpy as np
import onnx
import onnxslim
import torch
from xslim.logger import logger

from . import CalibrationCollect, XSlimDataset
from .analyse import statistical_analyse
from .defs import XQUANT_CONFIG
from .onnx_graph_helper import (format_onnx_model, merge_onnx_model,
                                truncate_onnx_model)
from .optimizer import GraphLegalized
from .ppq_decorator import (DISPATCHER_TABLE, BaseGraph, GraphDispatcher,
                            OnnxParser, ONNXRUNTIMExporter, TargetPlatform,
                            TorchExecutor)
from .quantizer import (XSlimQuantizer, convert_to_fp16_onnx_model,
                        dynamic_quantize_onnx_model)
from .xslim_setting import XSlimSetting, XSlimSettingFactory


def dispatch_graph(graph: BaseGraph, dispatcher: Union[str, GraphDispatcher] = "conservative") -> BaseGraph:
    quantizer = XSlimQuantizer(graph)
    if isinstance(dispatcher, str):
        dispatcher = dispatcher.lower()
        if dispatcher not in DISPATCHER_TABLE:
            raise ValueError(f'Can not found dispatcher type "{dispatcher}", check your input again.')
        dispatcher = DISPATCHER_TABLE[dispatcher](graph)
    else:
        if not isinstance(dispatcher, GraphDispatcher):
            raise TypeError(
                'Parameter "dispachter" of function ppq.api.dispatch_graph must be String or GraphDispatcher, '
                f"however {type(dispatcher)} was given."
            )
        dispatcher = dispatcher

    assert isinstance(dispatcher, GraphDispatcher)
    quant_types = quantizer.quant_operation_types
    dispatching_table = dispatcher.dispatch(
        graph=graph,
        quant_types=quant_types,
        quant_platform=TargetPlatform.UNSPECIFIED,
        fp32_platform=TargetPlatform.FP32,
        SOI_platform=TargetPlatform.SOI,
    )

    for operation in graph.operations.values():
        assert (
            operation.name in dispatching_table
        ), f"Internal Error, Can not find operation {operation.name} in dispatching table."
        operation.platform = dispatching_table[operation.name]
    return graph


def xslim_load_onnx_graph(
    file_or_model: Union[str, onnx.ModelProto], sim_en: bool = True, truncate_var_name: Sequence[str] = []
):
    if isinstance(file_or_model, onnx.ModelProto):
        onnx_model = file_or_model
    elif isinstance(file_or_model, str):
        onnx_model = onnx.load(file_or_model)
    else:
        raise TypeError("type of file_or_model error, {} .vs str or modelproto".format(type(file_or_model)))

    onnx_model = format_onnx_model(onnx_model, sim_en)
    onnx_model, truncate_left_graph, truncate_vars = truncate_onnx_model(onnx_model, truncate_var_name)

    graph = OnnxParser().build(onnx_model)
    return graph, truncate_left_graph, truncate_vars


def parse_xslim_config(file_or_dict: Union[str, dict]) -> XSlimSetting:
    config_dict = file_or_dict
    if isinstance(file_or_dict, str):
        with open(file_or_dict, "r") as fp:
            config_dict = json.load(fp)

    return XSlimSettingFactory.from_json(config_dict)


def quantize_onnx_model(
    path_or_config: Union[str, dict],
    input_onnx_model_or_path: Optional[Union[str, onnx.ModelProto]] = None,
    output_path: Optional[str] = None,
) -> onnx.ModelProto:
    """
    xslim model quantize api

    Args:
        path_or_config (Union[str, dict]): xslim config json file or config dict
        input_onnx_model_or_path (Optional[Union[str, onnx.ModelProto]], optional): input onnx model proto or path. Defaults to None.
        output_path (Optional[str], optional): output model path or output model dir. Defaults to None.

    Raises:
        RuntimeError: maybe set input_onnx_model_or_path error

    Returns:
        onnx.ModelProto: output model onnx proto
    """
    time_start = time.time()
    config_setting = parse_xslim_config(path_or_config)

    model_path = config_setting.model_parameters.onnx_model

    if input_onnx_model_or_path is not None:
        logger.info("using api input onnx model {}.".format(input_onnx_model_or_path))
        if isinstance(input_onnx_model_or_path, onnx.ModelProto):
            model_path = input_onnx_model_or_path
        elif os.path.exists(input_onnx_model_or_path):
            model_path = input_onnx_model_or_path
            config_setting.model_parameters.output_prefix = os.path.splitext(
                os.path.basename(input_onnx_model_or_path)
            )[0]
        else:
            raise RuntimeError("input_onnx_model_or_path set error.")

    if isinstance(output_path, str):
        logger.info("using api output path {}.".format(output_path))
        if os.path.isdir(output_path):
            config_setting.model_parameters.working_dir = output_path
        elif output_path[-5:] == ".onnx":
            config_setting.model_parameters.working_dir = os.path.dirname(output_path)
            config_setting.model_parameters.output_prefix = os.path.splitext(os.path.basename(output_path))[0]
        else:
            config_setting.model_parameters.working_dir = output_path

    config_setting.model_parameters.working_dir = os.path.realpath(config_setting.model_parameters.working_dir)

    output_prefix = config_setting.model_parameters.output_prefix
    working_dir = config_setting.model_parameters.working_dir
    calibration_step = config_setting.calibration_parameters.calibration_step
    calibration_device = config_setting.calibration_parameters.calibration_device

    if not os.path.exists(working_dir):
        logger.info("{} not existed and make new one.".format(working_dir))
        os.makedirs(working_dir)

    if isinstance(model_path, onnx.ModelProto):
        ori_onnx_model = model_path
    elif isinstance(model_path, str):
        ori_onnx_model = onnx.load(model_path)
    else:
        raise TypeError("type of file_or_model error, {} .vs str or modelproto".format(type(file_or_model)))

    if config_setting.quantization_parameters.precision_level.value >= 100:
        quant_onnx_model = format_onnx_model(ori_onnx_model, not config_setting.model_parameters.skip_onnxsim)
    elif config_setting.quantization_parameters.precision_level.value >= 4:
        if len(config_setting.quantization_parameters.ignore_op_types) > 0:
            logger.info(f"Ignoring op types: {config_setting.quantization_parameters.ignore_op_types}")
        if len(config_setting.quantization_parameters.ignore_op_names) > 0:
            logger.info(f"Ignoring op names: {config_setting.quantization_parameters.ignore_op_names}")
        quant_onnx_model = convert_to_fp16_onnx_model(
            model_path,
            config_setting.quantization_parameters.ignore_op_types,
            config_setting.quantization_parameters.ignore_op_names,
            not config_setting.model_parameters.skip_onnxsim,
        )
    elif config_setting.quantization_parameters.precision_level.value == 3:
        if len(config_setting.quantization_parameters.ignore_op_types) > 0:
            logger.info(f"Ignoring op types: {config_setting.quantization_parameters.ignore_op_types}")
        if len(config_setting.quantization_parameters.ignore_op_names) > 0:
            logger.info(f"Ignoring op names: {config_setting.quantization_parameters.ignore_op_names}")
        quant_onnx_model = dynamic_quantize_onnx_model(
            model_path,
            config_setting.quantization_parameters.ignore_op_types,
            config_setting.quantization_parameters.ignore_op_names,
            not config_setting.model_parameters.skip_onnxsim,
        )
    else:
        ppq_ir, truncate_left_graph, truncate_vars = xslim_load_onnx_graph(
            model_path,
            not config_setting.model_parameters.skip_onnxsim,
            config_setting.quantization_parameters.truncate_var_names,
        )

        GraphLegalized(ppq_ir)()

        ppq_ir = dispatch_graph(graph=ppq_ir, dispatcher="conservative")

        config_setting.calibration_parameters.check_input_parametres(ppq_ir)
        input_parametres = config_setting.calibration_parameters.input_parametres

        data_set = XSlimDataset(config_setting.calibration_parameters)
        calib_dataloader = torch.utils.data.DataLoader(data_set, batch_size=data_set.auto_batch_size)
        quantizer = XSlimQuantizer(ppq_ir)
        executor = TorchExecutor(graph=quantizer._graph, device=calibration_device)

        collate_fn = CalibrationCollect(input_parametres, calibration_device)

        single_graph_input_name = None
        dummy_input = None
        for k, v in ppq_ir.inputs.items():
            single_graph_input_name = k
        for data in calib_dataloader:
            data = collate_fn(data)
            if isinstance(data, torch.Tensor) and len(ppq_ir.inputs) == 1:
                if collate_fn is not None:
                    dummy_input = {single_graph_input_name: data}
            elif isinstance(data, dict):
                dummy_input = data
            else:
                raise TypeError(type(data))

        quantizer.quantize(
            inputs=dummy_input,
            calib_dataloader=calib_dataloader,
            executor=executor,
            xslim_setting=config_setting,
            calib_steps=calibration_step,
            collate_fn=collate_fn,
        )

        if config_setting.quantization_parameters.analysis_enable:
            try:
                test_dataloader = torch.utils.data.DataLoader(data_set, shuffle=True)
                report_path = os.path.join(working_dir, "{}_report.md".format(output_prefix))
                graphwise_analyse_results = statistical_analyse(
                    quantizer._graph,
                    calibration_device,
                    test_dataloader,
                    collate_fn,
                    steps=XQUANT_CONFIG.analyse_steps,
                    report_path=report_path,
                )
            except:
                logger.warning("quantize analysis failed and skiped.")

        quant_onnx_model = ONNXRUNTIMExporter().export(
            quantizer._graph,
        )

        quant_onnx_model = merge_onnx_model(quant_onnx_model, truncate_left_graph, truncate_vars)

    quant_onnx_model.metadata_props.extend(ori_onnx_model.metadata_props)
    quant_onnx_model.ir_version = max(9, ori_onnx_model.ir_version)
    onnx.save(quant_onnx_model, os.path.join(working_dir, "{}.onnx".format(output_prefix)))

    logger.info("quantization eplased time {:.2f} s".format(time.time() - time_start))
    return quant_onnx_model



#!/usr/bin/env python3
# Copyright (c) 2023 SpacemiT. All rights reserved.
from typing import Any, Callable, Dict, Iterable, List, Set, Tuple, Union

import numpy as np
import onnx
import torch
from xslim.defs import GLOBAL_FUNCTIONS_MAPPING, MIN_ONNX_OPSET_VERSION
from xslim.logger import logger

from ..ppq_decorator import (
    BaseGraph,
    DataType,
    GraphFormatter,
    GraphMerger,
    Operation,
    Opset,
    SearchableGraph,
    Variable,
    convert_any_to_torch_tensor,
)


class GraphLegalized:
    def __init__(self, graph) -> None:
        self._graph = graph
        self._merger = GraphMerger(self._graph)
        self._formatter = GraphFormatter(self._graph)

    def __call__(self) -> Any:
        self._formatter.remove_constant_input()
        self.remove_empty_sequence_input()
        self._formatter.convert_to_tensor()
        self.format_cast()
        self._formatter.format_parameter()
        self._merger.fuse_bias_add()
        self._merger.fuse_bn()
        self._formatter.format_slice()
        self._formatter.format_clip()
        self._formatter.format_pad()
        self._formatter.format_resize()
        self._formatter.remove_identity()
        self._formatter.delete_isolated()
        self.format_reshape_squeeze()
        self._merger.fuse_layernorm()
        self._merger.fuse_gelu()
        self.format_div()
        self._merger.fuse_bias_add()
        self.remove_dropout()
        self.format_ms_domain()
        self.fuse_mul_add()
        self.fuse_mul_add()
        # self.fuse_matmul_bias()

    def remove_empty_sequence_input(self) -> None:
        removing_ops = []
        for op in self._graph.operations.values():
            if op.type == "SequenceEmpty":
                removing_ops.append(op)

        for const_op in removing_ops:
            assert isinstance(const_op, Operation)
            dtype = DataType(const_op.attributes.get("dtype", DataType.FP32))
            constant_value = np.empty(0, dtype=DataType.to_numpy(dtype))
            output_var = const_op.outputs[0]
            output_var._is_parameter = True
            output_var.value = constant_value
            self._graph.remove_operation(removing_op=const_op)

    def fuse_matmul_bias(self):
        function_impl = BaseGraph(name="BatchMatMul")
        var_A = Variable(name="A", dtype=DataType.FP32)
        var_B = Variable(name="B", dtype=DataType.FP32)
        var_C = Variable(name="C", dtype=DataType.FP32)
        var_temp = Variable(name="MatMul_temp", dtype=DataType.FP32)
        var_Y = Variable(name="Y", dtype=DataType.FP32)
        function_impl.operations["MatMul"] = Operation(
            name="MatMul",
            op_type="MatMul",
            attributes={},
            inputs=[var_A, var_B],
            outputs=[var_temp],
        )
        function_impl.operations["Add"] = Operation(
            name="Add",
            op_type="Add",
            attributes={},
            inputs=[var_temp, var_C],
            outputs=[var_Y],
        )
        var_A.dest_ops.append(function_impl.operations["MatMul"])
        var_B.dest_ops.append(function_impl.operations["MatMul"])
        var_C.dest_ops.append(function_impl.operations["Add"])
        var_temp.dest_ops.append(function_impl.operations["Add"])
        var_temp.source_op = function_impl.operations["MatMul"]
        var_Y.source_op = function_impl.operations["Add"]
        for op in function_impl.operations.values():
            for in_var in op.inputs:
                function_impl.variables[in_var.name] = in_var
            for out_var in op.outputs:
                function_impl.variables[out_var.name] = out_var

        function_impl.inputs[var_A.name] = var_A
        function_impl.inputs[var_B.name] = var_B
        function_impl.inputs[var_C.name] = var_C
        function_impl.outputs[var_Y.name] = var_Y
        function_impl._detail["function_input"] = [var_A.name, var_B.name, var_C.name]
        function_impl._detail["function_output"] = [var_Y.name]
        function_impl._detail["function_domain"] = "spacemit_functions"
        function_impl._detail["function_opset_import"] = [{"domain": "", "version": MIN_ONNX_OPSET_VERSION}]
        function_impl._detail["function_attribute"] = ["transA", "transB", "transY"]

        add_function_impl = False
        for current_op in [_ for _ in self._graph.operations.values()]:
            if current_op.type != "MatMul":
                continue

            next_ops = self._graph.get_downstream_operations(current_op)
            if len(next_ops) != 1:
                continue
            if next_ops[0].type != "Add":
                continue

            fusing_op = next_ops[0]
            if fusing_op.num_of_parameter == 1:
                bias = fusing_op.parameters[0].value
                if bias.ndim in {0, 1}:
                    add_function_impl = True
                    self._graph.remove_operation(fusing_op, keep_coherence=True)
                    self._graph.create_variable(value=bias, is_parameter=True, dest_ops=[current_op])
                    current_op.type = "BatchMatMul"
                    current_op.attributes["domain"] = "spacemit_functions"
                    current_op.attributes["transA"] = 0
                    current_op.attributes["transB"] = 0
                    current_op.attributes["transY"] = 0
                    current_op.opset = Opset("spacemit_functions", version=1)

        if add_function_impl:
            self._graph._detail[GLOBAL_FUNCTIONS_MAPPING]["spacemit_functions.BatchMatMul"] = function_impl

    def format_cast(self):
        interested_ops = []
        for _, operation in self._graph.operations.items():
            assert isinstance(operation, Operation)
            if operation.type == "Cast":
                interested_ops.append(operation)
        for operation in interested_ops:
            assert isinstance(operation, Operation)
            assert "to" in operation.attributes
            if isinstance(operation.attributes["to"], np.dtype):
                operation.attributes["to"] = DataType.convert_from_numpy(operation.attributes["to"])

    def format_div(self):
        for op in self._graph.operations.values():
            if op.type == "Div" and op.inputs[1].is_parameter:
                if op.inputs[1].value.dtype in {torch.float32, torch.float64, torch.float16}:
                    op.type = "Mul"
                    op.inputs[1].value = 1 / op.inputs[1].value

    def format_reshape_squeeze(self):
        search_engine = SearchableGraph(graph=self._graph)
        paths = search_engine.path_matching(
            sp_expr=lambda x: x.type in {"Reshape"}
            and len(x.inputs) > 1
            and x.inputs[1].is_parameter
            and len(x.outputs[0].dest_ops) == 1
            and x.outputs[0].name not in self._graph.outputs,
            rp_expr=lambda x, y: False,
            ep_expr=lambda x: x.type in {"Squeeze"},
            direction="down",
        )

        for path in paths:
            path = path.tolist()
            assert len(path) == 2, "Oops seems we got something unexpected."
            reshape_op, squeeze_op = path
            assert isinstance(reshape_op, Operation) and isinstance(squeeze_op, Operation)
            reshape_size = reshape_op.inputs[1].value
            squeeze_axes = squeeze_op.attributes.get("axes", None)
            if len(squeeze_op.inputs) > 1 and squeeze_op.inputs[1].is_parameter:
                squeeze_axes = squeeze_op.inputs[1].value.numpy().tolist()

            if squeeze_axes is not None and all([reshape_size[axes] == 1 for axes in squeeze_axes]):
                new_shape = [int(s) for i, s in enumerate(reshape_size) if i not in squeeze_axes]

                reshape_op.outputs[0] = squeeze_op.outputs[0]
                squeeze_op.outputs[0].source_op = reshape_op
                reshape_op.inputs[1].value = convert_any_to_torch_tensor(
                    new_shape, device=reshape_size.device, dtype=reshape_size.dtype
                )

                for in_var in squeeze_op.inputs:
                    in_var.dest_ops.clear()
                    in_var.source_op = None
                    self._graph.remove_variable(in_var)

                squeeze_op.inputs.clear()
                squeeze_op.outputs.clear()
                self._graph.remove_operation(squeeze_op)

    def remove_dropout(self):
        removing_ops = []
        for op in self._graph.operations.values():
            if op.type == "Dropout":
                removing_ops.append(op)

        for op in removing_ops:
            in_var = op.inputs[0]
            out_var = op.outputs[0]

            in_var.dest_ops.remove(op)
            out_var.source_op = None

            for var in op.inputs[1:]:
                var.dest_ops.remove(op)
                self._graph.remove_variable(var)

            for var in op.outputs[1:]:
                var.source_op = None
                self._graph.remove_variable(var)

            op.inputs.clear()
            op.outputs.clear()

            for dest_op, dest_idx in zip([_ for _ in out_var.dest_ops], [_ for _ in out_var.dest_idx]):
                dest_op.inputs[dest_idx] = in_var
                in_var.dest_ops.append(dest_op)

            self._graph.remove_operation(op)
            self._graph.remove_variable(out_var)

    def format_ms_domain(self):
        has_ms_domain = False
        for op in self._graph.operations.values():
            if op.type in {"Gelu"}:
                op.attributes["domain"] = "com.microsoft"
                has_ms_domain = True
        if has_ms_domain:
            self._graph._detail["pb_opset_import"].append({"domain": "com.microsoft", "version": 1})

    def fuse_mul_add(self):
        search_engine = SearchableGraph(graph=self._graph)
        paths = search_engine.path_matching(
            sp_expr=lambda x: x.type in {"Conv", "Gemm", "ConvTranspose"} and x.num_of_parameter > 0,
            rp_expr=lambda x, y: False,
            ep_expr=lambda x: x.type in {"Mul", "Div", "Add", "Sub"}
            and any([in_var.is_parameter for in_var in x.inputs]),
            direction="down",
        )

        for path in paths:
            path = path.tolist()
            assert len(path) == 2, "Oops seems we got something unexpected."

            computing_op, mul_op = path
            assert isinstance(computing_op, Operation) and isinstance(mul_op, Operation)

            if (
                len(self._graph.get_downstream_operations(computing_op)) != 1
                or len(self._graph.get_upstream_operations(mul_op)) != 1
            ):
                continue

            parameter_index = [in_var.is_parameter for in_var in mul_op.inputs].index(True)
            feature_index = 1 - parameter_index
            parameter = mul_op.inputs[parameter_index]
            feature = mul_op.inputs[feature_index]

            if computing_op.num_of_parameter == 1:
                w = computing_op.parameters[0].value  # no bias.
                assert isinstance(w, torch.Tensor), "values of parameters are assumed as torch Tensor"
                if computing_op.type == "ConvTranspose":
                    b = torch.zeros(w.shape[1] * computing_op.attributes.get("group", 1)).to(w.device)
                elif computing_op.type == "Gemm" and computing_op.attributes.get("transB", 0) == 0:
                    b = torch.zeros(w.shape[1]).to(w.device)
                else:
                    b = torch.zeros(w.shape[0]).to(w.device)
            else:
                w, b = [var.value for var in computing_op.parameters[:2]]  # has bias.

            if parameter.value.numel() == 1 or parameter.value.ndim == w.ndim:
                pass
            else:
                continue

            if parameter.value.numel() == 1 or (
                parameter.value.ndim >= 2 and parameter.value.shape[1] == parameter.value.numel()
            ):
                pass
            else:
                continue

            if mul_op.type in {"Mul", "Div"}:
                alpha = parameter.value
                if alpha.dtype not in {torch.float32, torch.float64, torch.float16}:
                    continue
                if mul_op.type == "Div":
                    if parameter_index != 1:
                        continue
                    alpha = 1.0 / alpha

                if computing_op.type == "Conv":
                    # calculate new weight and bias
                    scale = alpha
                    w = w * scale.reshape([-1] + [1] * (w.ndim - 1))
                    b = alpha.reshape(-1) * b

                elif computing_op.type == "Gemm":
                    # calculate new weight and bias
                    scale = alpha
                    if computing_op.attributes.get("transB", 0):
                        w = w * scale.reshape([-1, 1])
                    else:
                        w = w * scale.reshape([1, -1])
                    b = alpha.reshape(-1) * b

                elif computing_op.type == "ConvTranspose":
                    scale = alpha
                    group = computing_op.attributes.get("group", 1)
                    scale = scale.reshape([group, 1, -1] + [1] * (w.ndim - 2))
                    w = w.reshape([group, -1] + list(w.shape[1:])) * scale
                    w = w.reshape([w.shape[0] * w.shape[1]] + list(w.shape[2:]))
                    b = alpha.reshape(-1) * b
                else:
                    raise TypeError(
                        f"Unexpected op type {computing_op.type}. "
                        f"Can not merge {computing_op.name} with {mul_op.name}"
                    )
            else:
                beta = parameter.value
                if mul_op.type == "Div":
                    if parameter_index != 1:
                        continue
                    beta = -beta
                b = beta.reshape(-1) + b

            # create new op and variable
            merged_op = Operation(
                computing_op.name, op_type=computing_op.type, attributes=computing_op.attributes.copy()
            )
            weight_var = Variable(computing_op.name + "_weight", w, True, [merged_op])
            bias_var = Variable(computing_op.name + "_bias", b, True, [merged_op])

            # replace & dirty work
            input_var = computing_op.inputs[0]
            output_var = mul_op.outputs[0]

            input_var.dest_ops.remove(computing_op)
            input_var.dest_ops.append(merged_op)

            output_var.source_op = merged_op

            # delete old operations
            computing_op.inputs.pop(0)
            mul_op.outputs.clear()
            self._graph.remove_operation(computing_op)
            self._graph.remove_operation(mul_op)

            # insert new
            self._graph.append_operation(merged_op)
            merged_op.inputs.extend([input_var, weight_var, bias_var])
            merged_op.outputs.extend([output_var])

            self._graph.append_variable(weight_var)
            self._graph.append_variable(bias_var)



#!/usr/bin/env python3
# Copyright (c) 2023 SpacemiT. All rights reserved.
import functools
import math
import os
from collections import OrderedDict
from datetime import datetime
from enum import Enum
from typing import Sequence, Set, Tuple, Union

import numpy as np
import onnx
import onnx_graphsurgeon as osg
from onnxruntime.quantization.quant_utils import \
    quantize_data as ort_quantize_data
from xslim.defs import XQUANT_CONFIG
from xslim.logger import logger

from ..onnx_graph_helper import format_onnx_model


def dynamic_weight_only_quantize(onnx_model, ignore_op_types_list, ignore_op_names_list, quant_bits=8):
    quant_min = -(2 ** (quant_bits - 1))
    quant_max = 2 ** (quant_bits - 1) - 1
    new_nodes = []

    def make_dq_node(weight, scale, zp, axis=0):
        dq_node_index = len(new_nodes)
        weight_tensor = osg.Constant(
            "insert_dq_node_weight_tensor_{}".format(dq_node_index), weight)
        scale_tensor = osg.Constant(
            "insert_dq_node_scale_tensor_{}".format(dq_node_index), scale)
        zp_tensor = osg.Constant(
            "insert_dq_node_zp_tensor_{}".format(dq_node_index), zp)
        out_tensor = osg.Variable(
            "insert_dq_node_out_tensor_{}".format(dq_node_index), np.float32)
        attrs = {}
        attrs["axis"] = axis
        new_node = osg.Node(
            "DequantizeLinear",
            "insert_dq_node_{}".format(dq_node_index),
            attrs,
            inputs=[weight_tensor, scale_tensor, zp_tensor],
            outputs=[out_tensor],
        )
        return new_node

    def make_dynamic_q_node(node, input_idx):
        dyn_q_node_index = len(new_nodes)
        dyn_quantize_node = osg.Node(
            "DynamicQuantizeLinear",
            "insert_dyn_q_node_{}_{}_{}".format(
                input_idx, node.name, dyn_q_node_index),
            inputs=[node.inputs[input_idx]],
            outputs=[],
        )

        dyn_quantize_node.outputs.append(
            osg.Variable("insert_dyn_q_node_out_{}".format(
                dyn_quantize_node.name), np.uint8)
        )
        dyn_quantize_node.outputs.append(
            osg.Variable("insert_dyn_q_node_out_scale_{}".format(
                dyn_quantize_node.name), np.float32)
        )
        dyn_quantize_node.outputs.append(
            osg.Variable("insert_dyn_q_node_out_zp_{}".format(
                dyn_quantize_node.name), np.uint8)
        )

        dyn_dequantize_node = osg.Node(
            "DequantizeLinear",
            "insert_dyn_dq_node_{}_{}_{}".format(
                input_idx, node.name, dyn_q_node_index),
            attrs={"axis": 0},
            inputs=dyn_quantize_node.outputs,
            outputs=[],
        )
        dyn_dequantize_node.outputs.append(
            osg.Variable("dyn_deq_node_out_{}".format(
                dyn_dequantize_node.name), np.float32)
        )
        return dyn_quantize_node, dyn_dequantize_node

    def get_scale_zp(weight_value):
        weight_scale_list = []
        weight_zp_list = []
        weight_quant_list = []
        for i in range(weight_value.shape[0]):
            weight_value_i = weight_value[i]
            # _, _, zero_point, scale, quantized_per_channel_data = ort_quantize_data(
            #     weight_value_i,
            #     onnx.TensorProto.INT8,
            #     True,
            #     False
            # )
            zero_point, scale, quantized_per_channel_data = ort_quantize_data(
                weight_value_i, onnx.TensorProto.INT8, True, False
            )
            weight_scale_list.append(scale)
            weight_zp_list.append(zero_point)
            weight_quant_list.append(quantized_per_channel_data)

        return np.array(weight_scale_list), np.array(weight_zp_list), np.array(weight_quant_list)

    def cosine_error(lhs_value, rhs_value):
        lhs_value = lhs_value.reshape(-1)
        rhs_value = rhs_value.reshape(-1)
        dot_product = np.sum(lhs_value * lhs_value)

        norm_a = np.linalg.norm(lhs_value, axis=-1)
        norm_b = np.linalg.norm(rhs_value, axis=-1)
        epsilon = 1e-10
        denominator = np.maximum(norm_a * norm_b, epsilon)
        similarity = dot_product / denominator
        return similarity

    osg_graph = osg.import_onnx(onnx_model)

    for node in osg_graph.nodes:

        def eval_error(weight_value, weight_q_value, weight_value_scale, weight_value_zp):
            requant_weight_value = (
                weight_q_value.astype(np.float32) -
                weight_value_zp.reshape(-1, 1)
            ) * weight_value_scale.reshape(-1, 1)
            error = np.sum(
                np.abs(weight_value - requant_weight_value)) / weight_value.size
            cosine = cosine_error(weight_value, requant_weight_value)
            # print(f"node: [{node.op}]({node.name}), error: {error}, cosine: {cosine}")

        if node.op in {"Conv", "Gemm", "MatMul", "ConvTranspose"}:
            if node.op in ignore_op_types_list:
                continue

            if node.name in ignore_op_names_list:
                continue

            axis = 0
            if isinstance(node.inputs[1], osg.Constant):
                weight_value = node.inputs[1].values
                weight_shape = weight_value.shape

                if np.allclose(
                    weight_value.astype(np.float32),
                    weight_value.astype(np.int32).astype(np.float32),
                    rtol=1e-03,
                    atol=1e-05,
                ):
                    continue

                if node.op == "Conv" or (node.op == "Gemm" and node.attrs.get("transB", 0) == 1):
                    weight_value = weight_value.reshape(weight_shape[0], -1)
                    weight_value_scale, weight_value_zp, quant_weight_value = get_scale_zp(
                        weight_value)
                    # eval_error(weight_value, quant_weight_value, weight_value_scale, weight_value_zp)
                    quant_weight_value = quant_weight_value.reshape(
                        weight_shape)
                elif node.op == "MatMul" or (node.op == "Gemm" and node.attrs.get("transB", 0) == 0):
                    axis = 1
                    permute_weight_value = np.transpose(weight_value, (1, 0))
                    permute_weight_value = permute_weight_value.reshape(
                        permute_weight_value.shape[0], -1)
                    weight_value_scale, weight_value_zp, permute_quant_weight_value = get_scale_zp(
                        permute_weight_value)
                    # eval_error(permute_weight_value, permute_quant_weight_value, weight_value_scale, weight_value_zp)
                    quant_weight_value = np.transpose(
                        permute_quant_weight_value, (1, 0)).reshape(weight_shape)
                elif node.op == "ConvTranspose" and node.attrs.get("group", 1) == 1:
                    axis = 1
                    group = node.attrs.get("group", 1)
                    weight_value = weight_value.reshape(
                        group, weight_shape[0] // group, weight_shape[1], -1)
                    permute_weight_value = np.transpose(
                        weight_value, (0, 2, 1, 3))
                    permute_weight_value = permute_weight_value.reshape(
                        group * weight_shape[1], -1)

                    weight_value_scale, weight_value_zp, permute_quant_weight_value = get_scale_zp(
                        permute_weight_value)
                    permute_quant_weight_value = permute_quant_weight_value.reshape(
                        (group, weight_shape[1], weight_shape[0] // group, -1)
                    )
                    quant_weight_value = np.transpose(
                        permute_quant_weight_value, (0, 2, 1, 3))
                    quant_weight_value = quant_weight_value.reshape(
                        weight_shape)
                else:
                    continue

                new_nodes.append(
                    make_dq_node(
                        quant_weight_value,
                        weight_value_scale.reshape(-1).astype(np.float32),
                        weight_value_zp.reshape(-1).astype(np.int8),
                        axis,
                    )
                )
                node.inputs[1] = new_nodes[-1].outputs[0]

    osg_graph.nodes.extend(new_nodes)
    osg_graph.toposort()
    new_onnx_model = osg.export_onnx(osg_graph)
    return new_onnx_model


def dynamic_quantize_onnx_model(
    file_or_model: Union[str, onnx.ModelProto],
    ignore_op_types_list: Sequence[str],
    ignore_op_names_list: Sequence[str],
    sim_en: bool = True,
):
    from onnxruntime.quantization import QuantizationMode, QuantType
    from onnxruntime.quantization.onnx_quantizer import ONNXQuantizer

    if isinstance(file_or_model, onnx.ModelProto):
        onnx_model = file_or_model
    elif isinstance(file_or_model, str):
        onnx_model = onnx.load(file_or_model)
    else:
        raise TypeError("type of file_or_model error, {} .vs str or modelproto".format(
            type(file_or_model)))

    onnx_model = format_onnx_model(onnx_model, sim_en)

    dynamic_q_op_types = {"Attention", "LSTM", "MatMul"}

    for ignore_op in ignore_op_types_list:
        if ignore_op in dynamic_q_op_types:
            dynamic_q_op_types.remove(ignore_op)

    nodes_to_exclude = ignore_op_names_list

    extra_options = {"WeightSymmetric": True, "MatMulConstBOnly": True,
                     "DefaultTensorType": onnx.TensorProto.FLOAT}
    quantizer = ONNXQuantizer(
        onnx_model,
        True,  # per channel
        False,  # reduce_range
        QuantizationMode.IntegerOps,
        False,  # static
        QuantType.QInt8,
        QuantType.QUInt8,
        None,
        [],
        nodes_to_exclude,
        list(dynamic_q_op_types),
        extra_options,
    )

    logger.info("quantize onnx model dynamic...")
    quantizer.quantize_model()
    quantized_model = quantizer.model.model
    quantized_model = dynamic_weight_only_quantize(
        quantized_model, ignore_op_types_list, ignore_op_names_list)
    quantized_model = format_onnx_model(quantized_model, True)

    quantized_model.producer_name = "xslim"
    export_time = quantized_model.metadata_props.add()
    export_time.key = "xslim_export_time"
    export_time.value = datetime.now().strftime("%Y-%m-%d-%H:%M:%S")
    xslim_version = quantized_model.metadata_props.add()
    xslim_version.key = "xslim_version"
    xslim_version.value = XQUANT_CONFIG.version

    return quantized_model



from .cosine import torch_cosine_similarity, numpy_cosine_similarity, torch_cosine_similarity_as_loss
from .statistic import torch_KL_divergence
from .norm import torch_mean_square_error, torch_snr_error



#!/usr/bin/env python3
# Copyright (c) 2023 SpacemiT. All rights reserved.

from . import ppq_decorator
from .analyse import statistical_analyse
from .calibration_helper import CalibrationCollect, XSlimDataset
from .defs import *
from .optimizer import *
from .quantizer import *
from .xslim_pipeline import quantize_onnx_model



import unittest
from pathlib import Path


REPO_ROOT = Path(__file__).resolve().parents[1]


class VersionMetadataTest(unittest.TestCase):
    def test_readme_version_matches_version_number(self):
        version = (REPO_ROOT / "VERSION_NUMBER").read_text(encoding="utf-8").strip()
        readme = (REPO_ROOT / "README.md").read_text(encoding="utf-8")
        readme_zh = (REPO_ROOT / "README_zh.md").read_text(encoding="utf-8")

        self.assertIn(f"version-{version}-blue.svg", readme)
        self.assertIn(f"| {version} | Current development version |", readme)
        self.assertIn(f"版本-{version}-blue.svg", readme_zh)
        self.assertIn(f"| {version} | 当前开发版本 |", readme_zh)

    def test_minimum_python_version_metadata_is_39(self):
        setup_py = (REPO_ROOT / "setup.py").read_text(encoding="utf-8")
        readme = (REPO_ROOT / "README.md").read_text(encoding="utf-8")
        readme_zh = (REPO_ROOT / "README_zh.md").read_text(encoding="utf-8")

        self.assertIn('python_requires=">=3.9"', setup_py)
        self.assertIn('Programming Language :: Python :: 3.9', setup_py)
        self.assertIn("python-%3E%3D3.9-blue.svg", readme)
        self.assertIn("python-%3E%3D3.9-blue.svg", readme_zh)



import onnx
import onnxslim

from xslim.logger import logger

from .dynamic_q_matmul import *
from .gelu import *
from .layernorm import *
from .padpool import *
from .batchnorm import *
# from .swish import *


def infer_onnx_model(onnx_model):
    opset_import = onnx_model.opset_import
    try:
        onnx_model = onnxslim.core.shape_infer(onnx_model)
    except Exception as e:
        logger.warning(
            f"onnx shape_inference error and skip. {e}")
    if onnx_model.opset_import is None or len(onnx_model.opset_import) == 0:
        onnx_model.opset_import.extend(opset_import)
    return onnx_model


def optimize_onnx_model(onnx_model):
    onnx_model = onnxslim.slim(onnx_model)
    return onnx_model



from typing import Iterable

import torch
from ...core import QuantizationProperty, QuantizationStates, QuantizationVisibility, ppq_warning
from ...executor import BaseGraphExecutor, TorchExecutor
from ...IR import BaseGraph, QuantableOperation
from ..observer import OperationObserver

from .base import QuantizationOptimizationPass


class PassiveParameterQuantizePass(QuantizationOptimizationPass):
    """
    ## PPQ Passive Parameter Quantization Pass(通用被动量化过程)

    Passive Parameters are those parameters that must share a same scale and offset with
    other tensors. This pass process 4 types of passive parameter by default, namely:

        Bias value in Gemm, it must has a scale = input scale * weight scale.

        Bias value in Conv, it must has a scale = input scale * weight scale.

        Clip min & Clip max in Clip, must has a scale = input scale

        Pading Value, must has a scale = input scale

    ### Parameters:

    * process_clip(Set[str]):

            Whether to process clip min, max

            If not processed, clip min, max will has their state = QuantizationState.FP32

    * process_bias(bool)

            Whether to process bias

            If not processed, bias will has their state = QuantizationState.ACTIVED

    * process_pad(bool)

            Whether to process clip min, max

            If not processed, pad value will has their state = QuantizationState.SOI

    * clip_visiblity(bool)

            Whether to export quant info of clip min, max

    * pad_visiblity(bool)

            Whether to export quant info of pad value

    ### Usage
    This pass is included in PPQ Quantization Setting, you can calling this optimization by:

        setting = QuantizationSettingFactory.default_setting()

        setting.parameter_setting.quantize_passive_parameter = True = True
        # calling ppq.api.quantize_onnx_model function with this setting.
        ir = quantize_torch_model(
        model=model, calib_dataloader=load_calibration_dataset(), setting=setting,
        platform=TargetPlatform.PPL_CUDA_INT8, calib_steps=8, input_shape=INPUT_SHAPE,
        collate_fn=collate_fn)
    """

    def __init__(
        self,
        process_clip: bool = True,
        process_bias: bool = True,
        process_pad: bool = True,
        clip_visiblity: QuantizationVisibility = QuantizationVisibility.INTERNAL,
        pad_visiblity: QuantizationVisibility = QuantizationVisibility.INTERNAL,
    ):
        self.process_clip = process_clip
        self.process_bias = process_bias
        self.process_pad = process_pad
        self.clip_visiblity = clip_visiblity
        self.pad_visiblity = pad_visiblity
        super().__init__(name="PPQ Passive Parameter Quantization")

    def optimize(self, graph: BaseGraph, **kwargs) -> None:
        def check_state(state: QuantizationStates):
            return state in {
                QuantizationStates.PASSIVE,
                QuantizationStates.ACTIVATED,
                QuantizationStates.BAKED,
                QuantizationStates.OVERLAPPED,
            }

        for op in graph.operations.values():
            if not isinstance(op, QuantableOperation):
                continue

            if op.type in {"Conv", "ConvTranspose", "Gemm"} and self.process_bias:
                # inputs are [input value, weight, bias(optional)]
                if op.num_of_input == 3:
                    i_cfg, w_cfg, b_cfg = op.config.input_quantization_config
                    if b_cfg.state not in {QuantizationStates.PASSIVE, QuantizationStates.PASSIVE_INIT}:
                        continue

                    # PATCH 2022.07.29 有的时候 bias 是个多维的东西，此时要求前面的维度都是1
                    bias = op.inputs[-1].value
                    if bias is None:
                        raise ValueError(
                            f"Bias Varaible {op.inputs[-1].name} must be a constant. " "Please check it again."
                        )

                    assert bias.numel() == bias.shape[-1], (
                        f"For op {op.name}, expect Bias shape to be {[bias.numel()]}, "
                        f"however {bias.shape} was given"
                    )
                    op.inputs[-1].value = bias.squeeze()
                    # PATCH 2022.08.02 只有一个数的 bias 经过 squeeze 会变成零维的, 再给它多加一维补回来
                    if op.inputs[-1].value.ndim == 0 and op.inputs[-1].value.numel() == 1:
                        op.inputs[-1].value = op.inputs[-1].value.unsqueeze(0)

                    if not check_state(i_cfg.state):
                        raise PermissionError(
                            f"Can not quantize bias of layer {op.name}, "
                            "cause input has not been correctly quantized."
                        )

                    b_cfg.scale = w_cfg.scale * i_cfg.scale
                    b_cfg.state = QuantizationStates.PASSIVE
                    b_cfg.offset = torch.zeros_like(b_cfg.scale)
                    assert not b_cfg.policy.has_property(
                        QuantizationProperty.ASYMMETRICAL
                    ), "Passive parameter does not support ASYMMETRICAL quantization"

            if op.type in {"Clip"} and self.process_clip:
                # inputs are [input value, min[optional], max[optional]]
                i_cfg = op.config.input_quantization_config[0]

                if not check_state(i_cfg.state):
                    raise PermissionError(
                        f"Can not quantize clip value of layer {op.name}, "
                        "cause input has not been correctly quantized."
                    )

                for config in op.config.input_quantization_config[1:]:
                    config.master_by = i_cfg
                    config.visibility = self.clip_visiblity

            if op.type in {"Pad"} and self.process_pad:
                # inputs are [input value, pad[shape-related], pad value[optional]]
                if op.num_of_input != 3:
                    continue
                i_cfg = op.config.input_quantization_config[0]

                if not check_state(i_cfg.state):
                    raise PermissionError(
                        f"Can not quantize pad value of layer {op.name}, "
                        "cause input has not been correctly quantized."
                    )

                if len(op.config.input_quantization_config) > 1:
                    pad_config = op.config.input_quantization_config[-1]
                    pad_config.master_by = i_cfg
                    pad_config.visibility = self.pad_visiblity

        # final check
        for op in graph.operations.values():
            if not isinstance(op, QuantableOperation):
                continue
            for cfg, var in op.config_with_variable:
                if cfg.state == QuantizationStates.PASSIVE_INIT:
                    ppq_warning(
                        f"Unexpected quantization state of variable {var.name} at op {op.name}, "
                        "The configuration state has been initialized as PASSIVE_INIT, "
                        "however PassiveParameterQuantizePass do not kown how to deal with it."
                    )


class ParameterQuantizePass(QuantizationOptimizationPass):
    """PPQ PassiveParameterQuantizePass completes the quantization of positive
    parameters. By default, all parameters with initial state will be quantized
    during this optimization, all non-parameter tensors will be excluded from
    this pass by temporary dequantization.

    Then, operation observers will be established automatically to record necessary statistics,
        observers are also responsible for rendering quantization configuration (computing scale and offset).

    This pass needs no data, however it uses fake data to finish a dummy forward process.
    see also: TorchExecutor.dummy_forward function
    """

    def __init__(self, method: str = None):
        self._method = method
        super().__init__(name="PPQ Parameter Quantization Pass")

    def optimize(self, graph: BaseGraph, dataloader: Iterable, executor: TorchExecutor, **kwargs) -> None:
        # build observer and hook for each quantable operation
        hooks, observers, state_records = {}, {}, {}
        for op_name, operation in graph.operations.items():
            if not isinstance(operation, QuantableOperation):
                continue

            for config, var in operation.config_with_variable:
                # deactivate non-parameter variable quantization just for now
                if not var.is_parameter:
                    state_records[config] = config.state
                    config.state = QuantizationStates.FP32
                elif self._method is not None:
                    # override quantizer's setting if necessary
                    config.observer_algorithm = self._method

            observer = OperationObserver(
                operation=executor._graph.operations[op_name], monitor_outputs=False, monitor_inputs=False
            )
            observers[op_name] = observer
            hooks[op_name] = observer.hook

        # dummy forward, quant all parameter.
        assert isinstance(executor, TorchExecutor), "ParameterQuantizePass Only support TorchExecutor now."
        executor.dummy_forward(hooks=hooks)

        # render quantization config, restore non-parameter quantization state
        for op_name, operation in graph.operations.items():
            if not isinstance(operation, QuantableOperation):
                continue

            for cfg, var in operation.config_with_variable:
                if not var.is_parameter:
                    cfg.state = state_records[cfg]

            observer = observers[op_name]
            assert isinstance(observer, OperationObserver)
            observer.render_quantization_config()
            observer.report()



import numpy as np
import onnxslim.third_party.onnx_graphsurgeon as osg
from onnxslim.core.pattern import Pattern, PatternMatcher
from onnxslim.core.pattern.registry import register_fusion_pattern


def _get_constant_pad_value(pad_node):
    if len(pad_node.inputs) < 3:
        return 0.0

    value_input = pad_node.inputs[2]
    if isinstance(value_input, osg.Variable) and value_input.name == "":
        return 0.0

    if not isinstance(value_input, osg.Constant):
        return None

    values = np.asarray(value_input.values)
    if values.size == 0:
        return 0.0

    if not np.all(values == values.reshape(-1)[0]):
        return None

    return values.reshape(-1)[0].item()


def _has_supported_axes_input(pad_node):
    if len(pad_node.inputs) < 4:
        return True

    axes_input = pad_node.inputs[3]
    return isinstance(axes_input, osg.Variable) and axes_input.name == ""


def _extract_pool_pads(pad_node):
    if not isinstance(pad_node.inputs[1], osg.Constant):
        return None

    pad_values = np.asarray(pad_node.inputs[1].values).reshape(-1).tolist()
    if len(pad_values) < 6 or len(pad_values) % 2 != 0:
        return None

    rank = len(pad_values) // 2
    if any(pad != 0 for pad in (pad_values[:2] + pad_values[rank: rank + 2])):
        return None

    return [int(pad) for pad in (pad_values[2:rank] + pad_values[rank + 2:])]


def _get_existing_pool_pads(pool_node, spatial_rank):
    auto_pad = pool_node.attrs.get("auto_pad", "NOTSET")
    if auto_pad not in {"NOTSET", "VALID"}:
        return None

    existing_pads = list(pool_node.attrs.get("pads", [0] * (2 * spatial_rank)))
    if len(existing_pads) != 2 * spatial_rank:
        return None

    if any(pad != 0 for pad in existing_pads):
        return None

    return existing_pads


def _rewrite_pool_with_input(pool_node, pad_node, merged_pads):
    input_variable = pad_node.inputs[0]
    pad_variable = pad_node.outputs[0]
    pool_node.inputs[pool_node.inputs.index(pad_variable)] = input_variable

    attrs = dict(pool_node.attrs)
    attrs["pads"] = merged_pads
    attrs.pop("auto_pad", None)

    inputs = list(pool_node.inputs)
    outputs = list(pool_node.outputs)

    pool_node.inputs.clear()
    pool_node.outputs.clear()

    if len(pad_node.users) == 0:
        if pad_node in input_variable.outputs:
            input_variable.outputs.remove(pad_node)
        pad_node.inputs.clear()
        pad_node.outputs.clear()

    return {
        pool_node.name: {
            "op": pool_node.op,
            "inputs": inputs,
            "outputs": outputs,
            "name": pool_node.name,
            "attrs": attrs,
            "domain": None,
        }
    }


class PadAveragePoolMatcher(PatternMatcher):
    def __init__(self, priority):
        pattern = Pattern(
            """
            input       input   0 1 pad_0
            Pad         pad_0   1+ 1 input pool_0
            AveragePool pool_0  1+ 1 pad_0 output
            output      output  1 0 pool_0
            """
        )
        super().__init__(pattern, priority)

    @property
    def name(self):
        return "FusionPadAveragePool"

    def parameter_check(self) -> bool:
        pad_node = getattr(self, "pad_0")
        is_constant_pad = isinstance(pad_node.inputs[1], osg.Constant)
        return is_constant_pad and _has_supported_axes_input(pad_node)

    def rewrite(self, opset=11):
        pad_node = getattr(self, "pad_0")
        pool_node = getattr(self, "pool_0")

        if pad_node.attrs.get("mode", "constant") != "constant":
            return {}

        pad_fill_value = _get_constant_pad_value(pad_node)
        if pad_fill_value is None or not np.isclose(pad_fill_value, 0.0):
            return {}

        pool_pads = _extract_pool_pads(pad_node)
        if pool_pads is None:
            return {}

        existing_pads = _get_existing_pool_pads(pool_node, len(pool_pads) // 2)
        if existing_pads is None:
            return {}

        merged_pads = [a + b for a, b in zip(pool_pads, existing_pads)]
        match_case = _rewrite_pool_with_input(pool_node, pad_node, merged_pads)
        match_case[pool_node.name]["attrs"]["count_include_pad"] = 1
        return match_case


class PadMaxPoolMatcher(PatternMatcher):
    def __init__(self, priority):
        pattern = Pattern(
            """
            input   input   0 1 pad_0
            Pad     pad_0   1+ 1 input pool_0
            MaxPool pool_0  1+ 1 pad_0 output
            output  output  1 0 pool_0
            """
        )
        super().__init__(pattern, priority)

    @property
    def name(self):
        return "FusionPadMaxPool"

    def parameter_check(self) -> bool:
        pad_node = getattr(self, "pad_0")
        is_constant_pad = isinstance(pad_node.inputs[1], osg.Constant)
        return is_constant_pad and _has_supported_axes_input(pad_node)

    def rewrite(self, opset=11):
        pad_node = getattr(self, "pad_0")
        pool_node = getattr(self, "pool_0")

        if pad_node.attrs.get("mode", "constant") != "constant":
            return {}

        pad_fill_value = _get_constant_pad_value(pad_node)
        if pad_fill_value is None or not np.isneginf(pad_fill_value):
            return {}

        pool_pads = _extract_pool_pads(pad_node)
        if pool_pads is None:
            return {}

        existing_pads = _get_existing_pool_pads(pool_node, len(pool_pads) // 2)
        if existing_pads is None:
            return {}

        merged_pads = [a + b for a, b in zip(pool_pads, existing_pads)]
        return _rewrite_pool_with_input(pool_node, pad_node, merged_pads)


register_fusion_pattern(PadAveragePoolMatcher(1))
register_fusion_pattern(PadMaxPoolMatcher(1))



from decimal import ROUND_HALF_DOWN, ROUND_HALF_EVEN, ROUND_HALF_UP, Decimal
from math import ceil, floor, log2
from typing import Union

import torch

from torch.autograd import Function
from ..core import RoundingPolicy


class PPQTensorRoundImpl(Function):
    @staticmethod
    def forward(ctx, value: torch.Tensor, policy: RoundingPolicy = RoundingPolicy.ROUND_HALF_EVEN) -> torch.Tensor:
        """
            reference: https://en.wikipedia.org/wiki/Rounding

        Args:
            value (torch.Tensor): [description]
            policy (RoundingPolicy, optional): [description]. Defaults to RoundingPolicy.ROUND_HALF_EVEN.

        Raises:
            ValueError: [description]

        Returns:
            torch.Tensor: [description]
        """
        assert isinstance(value, torch.Tensor), "tensor round only takes effect on torch tensor."
        if policy == RoundingPolicy.ROUND_HALF_EVEN:
            # default rounding policy of torch is ROUND_TO_NEAR_EVEN
            # try this: print(torch.Tensor([1.5, 2.5, 3.5, 4.5]).round())
            # However it may generate unexpected results due to version difference.
            return value.round()
        elif policy == RoundingPolicy.ROUND_UP:
            return value.ceil()
        elif policy == RoundingPolicy.ROUND_HALF_TOWARDS_ZERO:
            return torch.sign(value) * torch.ceil(value.abs() - 0.5)
        elif policy == RoundingPolicy.ROUND_HALF_FAR_FORM_ZERO:
            return torch.sign(value) * torch.floor(value.abs() + 0.5)
        elif policy == RoundingPolicy.ROUND_HALF_DOWN:
            return torch.ceil(value - 0.5)
        elif policy == RoundingPolicy.ROUND_HALF_UP:
            return torch.floor(value + 0.5)
        elif policy == RoundingPolicy.ROUND_TO_NEAR_INT:
            raise NotImplementedError(
                f"Torch Tensor can not use this rounding policy({policy}) try ROUND_HALF_EVEN instead."
            )
        else:
            raise ValueError("Unexpected rounding policy found.")

    @staticmethod
    def backward(ctx, dy: torch.Tensor):
        return dy, None


def ppq_numerical_round(value: float, policy: RoundingPolicy = RoundingPolicy.ROUND_HALF_EVEN) -> int:
    """
        reference: https://en.wikipedia.org/wiki/Rounding

        decimal definition:
            - decimal.ROUND_CEILING (towards Infinity)
            - decimal.ROUND_DOWN (towards zero)
            - decimal.ROUND_FLOOR (towards -Infinity)
            - decimal.ROUND_HALF_DOWN (to nearest with ties going towards zero)
            - decimal.ROUND_HALF_EVEN (to nearest with ties going to nearest even integer)
            - decimal.ROUND_HALF_UP (to nearest with ties going away from zero)
            - decimal.ROUND_UP (away from zero)
            - decimal.ROUND_05UP (away from zero if last digit after rounding towards zero would have been 0 or 5; otherwise towards zero)

    Args:
        value (float): [description]
        policy (RoundingPolicy, optional): [description]. Defaults to RoundingPolicy.ROUND_HALF_EVEN.

    Raises:
        ValueError: [description]

    Returns:
        int: [description]
    """
    assert isinstance(value, float), "numerical round only takes effect on float number."
    if policy == RoundingPolicy.ROUND_HALF_EVEN:
        return int(Decimal(value).quantize(exp=Decimal(1), rounding=ROUND_HALF_EVEN))
    elif policy == RoundingPolicy.ROUND_HALF_UP:
        if value > 0:
            return int(Decimal(value).quantize(exp=Decimal(1), rounding=ROUND_HALF_UP))
        else:
            return int(Decimal(value).quantize(exp=Decimal(1), rounding=ROUND_HALF_DOWN))
    elif policy == RoundingPolicy.ROUND_HALF_DOWN:
        if value > 0:
            return int(Decimal(value).quantize(exp=Decimal(1), rounding=ROUND_HALF_DOWN))
        else:
            return int(Decimal(value).quantize(exp=Decimal(1), rounding=ROUND_HALF_UP))
    elif policy == RoundingPolicy.ROUND_HALF_TOWARDS_ZERO:
        return ppq_numerical_round(value, RoundingPolicy.ROUND_HALF_DOWN)
    elif policy == RoundingPolicy.ROUND_HALF_FAR_FORM_ZERO:
        return ppq_numerical_round(value, RoundingPolicy.ROUND_HALF_UP)
    elif policy == RoundingPolicy.ROUND_TO_NEAR_INT:
        if value > 0:
            return floor(value + 0.5)
        else:
            return ceil(value - 0.5)
    elif policy == RoundingPolicy.ROUND_UP:
        return ceil(value)
    else:
        raise ValueError("Unexpected rounding policy found.")


def ppq_tensor_round(value: torch.Tensor, policy: RoundingPolicy = RoundingPolicy.ROUND_HALF_EVEN) -> torch.Tensor:
    """
        reference: https://en.wikipedia.org/wiki/Rounding

    Args:
        value (torch.Tensor): [description]
        policy (RoundingPolicy, optional): [description]. Defaults to RoundingPolicy.ROUND_HALF_EVEN.

    Raises:
        ValueError: [description]

    Returns:
        torch.Tensor: [description]
    """
    return PPQTensorRoundImpl.apply(value, policy)


def ppq_round_to_power_of_2(value: Union[float, int], policy: RoundingPolicy = RoundingPolicy.ROUND_UP) -> float:
    """
    Round a given value to Integer, under Power-of-2 restrction.

    ### Formula:
    ppq_round_to_power_of_2(x) = sign * float(pow(2, round(log2(sign * value), policy=policy)))
        where sign is +1 or -1

    Args:
        value (Union[float, int]): _description_
        policy (RoundingPolicy, optional): _description_. Defaults to RoundingPolicy.ROUND_UP.

    Returns:
        float: _description_
    """
    if value == 0:
        return 0
    sign = 1 if value >= 0 else -1
    assert isinstance(value, float) or isinstance(value, int), "power-of-2 round only takes effect on float or int."
    return sign * float(pow(2, ppq_numerical_round(log2(sign * value), policy=policy)))



#!/usr/bin/env python3
# Copyright (c) 2023 SpacemiT. All rights reserved.
import functools
import importlib
import os
import sys
from collections import OrderedDict
from typing import Any, Callable, Dict, Sequence, Tuple, Union

import cv2
import numpy as np
import torch
import torchvision.transforms as transforms
from PIL import Image
from torch.utils.data import Dataset

from .xslim_setting import CalibrationParameterSetting, InputParameterSetting


class XSlimDataset(Dataset):
    def __init__(self, calibration_parameters: CalibrationParameterSetting):
        self.calibration_parameters = calibration_parameters
        input_parametres = self.calibration_parameters.input_parametres

        self._data_list = []
        self._data_dict = OrderedDict()
        min_size = 100000
        for input_item in input_parametres:
            data_list_path = input_item.data_list_path
            input_name = input_item.input_name
            if isinstance(data_list_path, str):
                relate_dir = os.path.dirname(data_list_path)
                with open(data_list_path, "r") as fp:
                    data_list_item = fp.readlines()
                    self._data_dict[input_name] = [
                        os.path.join(relate_dir, item.strip()) for item in data_list_item if len(item.strip()) > 0
                    ]
            else:
                self._data_dict[input_name] = [torch.zeros(input_item.input_shape).to(getattr(torch, input_item.dtype))]

            assert len(self._data_dict[input_name]), "Calibration input {} finds 0 items.".format(input_name)

            # 超过1个输入可以广播
            if len(self._data_dict[input_name]) == 1 and len(input_parametres) > 1:
                pass
            else:
                min_size = min(len(self._data_dict[input_name]), min_size)

        self.auto_batch_size = (
            input_parametres[0].input_shape[0]
            if len(input_parametres) == 1 and input_parametres[0].file_type == "img"
            else 1
        )

        self.auto_batch_size = (
            calibration_parameters.calibration_batch_size
            if calibration_parameters.calibration_batch_size > 1
            else self.auto_batch_size
        )

        min_size = (min_size // self.auto_batch_size) * self.auto_batch_size

        for i in range(min_size):
            data_iter = OrderedDict()
            for k, v in self._data_dict.items():
                data_iter[k] = v[0] if len(v) == 1 else v[i]
            self._data_list.append(data_iter)

    def __len__(self):
        return len(self._data_list)

    def __getitem__(self, idx):
        return self._data_list[idx]


class PTImagenetPreprocess:
    def __init__(self, out_height, out_width, mean_value, std_value):
        self.out_height = out_height
        self.out_width = out_width
        self.mean_value = mean_value
        self.std_value = std_value

    def __call__(self, img):
        import torchvision.transforms.functional as F

        img = Image.fromarray(img)
        img = F.resize(img, int(self.out_height / 0.875), Image.BILINEAR)
        img = F.center_crop(img, self.out_height)
        img = F.pil_to_tensor(img).to(torch.float32)
        img = F.normalize(img, mean=self.mean_value, std=self.std_value, inplace=False)
        return img


class ImagenetPreprocess:
    def resize_with_aspectratio(
        self, img: np.ndarray, out_height: int, out_width: int, scale: float = 87.5, inter_pol=cv2.INTER_LINEAR
    ):
        height, width, _ = img.shape
        new_height = int(100.0 * out_height / scale)
        new_width = int(100.0 * out_width / scale)
        if height > width:
            w = new_width
            h = int(new_height * height / width)
        else:
            h = new_height
            w = int(new_width * width / height)
        img = cv2.resize(img, (w, h), interpolation=inter_pol)
        return img

    def center_crop(self, img: np.ndarray, out_height: int, out_width: int):
        height, width, _ = img.shape
        left = int((width - out_width) / 2)
        right = int((width + out_width) / 2)
        top = int((height - out_height) / 2)
        bottom = int((height + out_height) / 2)
        img = img[top:bottom, left:right]
        return img

    def __init__(self, out_height: int, out_width: int, mean_value, std_value):
        self.out_height = out_height
        self.out_width = out_width
        self.mean_value = mean_value
        self.std_value = std_value

    def __call__(self, img):
        cv2_interpol = cv2.INTER_AREA
        img = self.resize_with_aspectratio(img, self.out_height, self.out_width, inter_pol=cv2_interpol)
        img = self.center_crop(img, self.out_height, self.out_width)
        img = np.asarray(img, dtype="float32")
        means = np.array(self.mean_value, dtype=np.float32) if self.mean_value is not None else 0
        stds = np.array(self.std_value, dtype=np.float32) if self.std_value is not None else 1.0
        img = (img - means) / stds
        img = img.transpose([2, 0, 1])
        return torch.from_numpy(img)


class CalibrationCollect:
    def __init__(self, input_parametres: Sequence[InputParameterSetting], calibration_device: str = "cuda") -> None:
        self.input_parametres = input_parametres
        self.calibration_device = calibration_device

        self.input_info_dict: Dict[str, InputParameterSetting] = OrderedDict()
        self.imagenet_transforms: Dict[str, Callable] = OrderedDict()
        self.custom_transforms: Dict[str, Callable] = OrderedDict()

        for input_info in self.input_parametres:
            self.input_info_dict[input_info.input_name] = input_info

            input_shape = input_info.input_shape
            mean_value = input_info.mean_value
            std_value = input_info.std_value
            calib_dataset_type = input_info.preprocess_file
            self.custom_transforms[input_info.input_name] = None
            self.imagenet_transforms[input_info.input_name] = None
            if calib_dataset_type == "PT_IMAGENET":
                self.imagenet_transforms[input_info.input_name] = transforms.Compose(
                    [PTImagenetPreprocess(input_shape[-2], input_shape[-1], mean_value, std_value)]
                )
            elif calib_dataset_type == "IMAGENET":
                self.imagenet_transforms[input_info.input_name] = transforms.Compose(
                    [ImagenetPreprocess(input_shape[-2], input_shape[-1], mean_value, std_value)]
                )
            elif isinstance(calib_dataset_type, str):
                module_file, func_name = calib_dataset_type.split(":")
                if os.path.dirname(module_file) not in sys.path:
                    sys.path.append(os.path.dirname(module_file))
                module_name = os.path.splitext(os.path.basename(module_file))[0]

                custom_module = importlib.import_module(module_name)
                self.custom_transforms[input_info.input_name] = getattr(custom_module, func_name)

    def __call__(self, data_item) -> Any:
        @functools.lru_cache(maxsize=(1024 * 1024 * 256))
        def calibration_data_collect(
            collector: CalibrationCollect,
            file_names: Tuple[str],
            tensor_name: str,
        ) -> torch.Tensor:
            input_info = collector.input_info_dict[tensor_name]
            if isinstance(collector.custom_transforms[tensor_name], Callable):
                return collector.custom_transforms[tensor_name](file_names, input_info.to_dict()).to(
                    getattr(torch, input_info.dtype)
                )

            file_type = input_info.file_type
            file_type = input_info.file_type
            color_format = input_info.color_format
            input_shape = input_info.input_shape
            mean_value = np.array(input_info.mean_value) if input_info.mean_value is not None else 0
            std_value = np.array(input_info.std_value) if input_info.std_value is not None else 1
            batch_list = []
            for batch_item in file_names:
                if isinstance(batch_item, torch.Tensor):
                    batch_list.append(batch_item)
                    continue
                elif file_type == "img":
                    if collector.imagenet_transforms[k] is not None:
                        img = cv2.imread(batch_item)
                        if color_format == "rgb":
                            img = cv2.cvtColor(img, cv2.COLOR_BGR2RGB)
                        img = collector.imagenet_transforms[k](img)
                        img = torch.unsqueeze(img, 0)
                        batch_list.append(img)
                    else:
                        img = cv2.imread(batch_item)
                        if len(img.shape) < 3:
                            # 灰度图
                            img = np.expand_dims(img, -1)
                        elif color_format == "rgb":
                            img = cv2.cvtColor(img, cv2.COLOR_BGR2RGB)
                        img = cv2.resize(img, (input_shape[-1], input_shape[-2]), interpolation=cv2.INTER_AREA)
                        img = img.astype(np.float32)
                        img = (img - mean_value) / std_value
                        img = np.transpose(img, (2, 0, 1))
                        img = torch.unsqueeze(torch.from_numpy(img), 0)
                        batch_list.append(img)
                elif file_type == "npy":
                    img = np.load(batch_item)
                    batch_list.append(torch.from_numpy(img))
                elif file_type == "raw":
                    img = np.fromfile(batch_item, dtype=np.float32)
                    batch_list.append(torch.from_numpy(img))
                else:
                    raise NotImplementedError("Calibration file type {}".format(file_type))

            return torch.cat(batch_list, dim=0).to(getattr(torch, input_info.dtype))

        dst_data_item = {}
        for k, v in data_item.items():
            dst_data_item[k] = calibration_data_collect(self, tuple(v), k)
        return dst_data_item



class PPQ_GLOBAL_CONFIGURATION:
    def __init__(self) -> None:
        # 是否启动 cuda kernel 加速计算
        self.USING_CUDA_KERNEL        = False
        
        # PPQ 的名字
        self.NAME                     = 'PPL Quantization Tool'
        
        # PPQ 的版本号
        self.VERSION                  = '0.6.6'
        
        # 导出图时是否导出权重（仅影响 Native 格式导出）
        self.DUMP_VALUE_WHEN_EXPORT   = True

        # 导出图时，是否导出调度信息
        self.EXPORT_PPQ_INTERNAL_INFO = False
        
        # 开启 PPQ 调试模式，将打印所有量化点插入信息
        self.PPQ_DEBUG                = False

PPQ_CONFIG = PPQ_GLOBAL_CONFIGURATION()  



import torch
from torch.autograd import Function
from ...core import PPQ_CONFIG, QuantizationProperty, QuantizationStates, RoundingPolicy, TensorQuantizationConfig


class TensorwiseFloatingQuantImpl(Function):
    """Torch Tensorwise quantize is designed to quantize a torch Tensor
    with a given configuration. All quantization within PPQ will invoke
    this function to quantize its value. Any modification of this function
    will greatly affects system behaviour.

    This is a torch implementation of quantization itself.
    Notice that if ppq.config.USING_CUDA_KERNAL = True,
        then all quantization will use ffi.CUDA instead.

    Notice this function will always clone your tensor value first.
    This function never quantize your tensor value inplace.
    """

    @staticmethod
    def forward(
        ctx,
        tensor: torch.Tensor,
        scales: torch.Tensor,
        offsets: torch.Tensor,
        exponet_bits: int,
        mantissa_bits: int,
        quant_min: float,
        quant_max: float,
        rounding: RoundingPolicy,
    ) -> torch.Tensor:
        scales, offsets = scales.to(tensor.device), offsets.to(tensor.device)
        if not PPQ_CONFIG.USING_CUDA_KERNEL or not tensor.is_cuda:
            # quantization function, pytorch implementation
            raise NotImplementedError("This Feature must run with PPQ Cuda Kernel.")

        else:
            raise NotImplementedError()

    @staticmethod
    def backward(ctx, dy: torch.Tensor):
        return dy, None, None, None, None, None, None, None, None


class ChannelwiseFloatingQuantImpl(Function):
    """Torch Channelwise quantize is designed to quantize a torch Tensor
    with a given configuration. All quantization within PPQ will invoke
    this function to quantize its value. Any modification of this function
    will greatly affects system behaviour.

    This is a torch implementation of quantization itself.
    Notice that if ppq.config.USING_CUDA_KERNAL = True,
        then all quantization will bypass this function by using ffi.CUDA instead.

    Notice this function will always clone your tensor value first.
    This function never quantize your tensor value inplace.
    """

    @staticmethod
    def forward(
        ctx,
        tensor: torch.Tensor,
        scales: torch.Tensor,
        offsets: torch.Tensor,
        channel_axis: int,
        exponet_bits: int,
        mantissa_bits: int,
        quant_min: float,
        quant_max: float,
        rounding: RoundingPolicy,
    ) -> torch.Tensor:
        scales, offsets = scales.to(tensor.device), offsets.to(tensor.device)
        if not PPQ_CONFIG.USING_CUDA_KERNEL or not tensor.is_cuda:
            # generate a shape that likes [1, 1, -1, 1], the only -1 is at channel axe.
            raise NotImplementedError("This Feature must run with PPQ Cuda Kernel.")
        else:
            raise NotImplementedError()

    @staticmethod
    def backward(ctx, dy: torch.Tensor):
        return dy, None, None, None, None, None, None, None, None, None


def PPQFloatingQuantFunction(tensor: torch.Tensor, config: TensorQuantizationConfig) -> torch.Tensor:
    if not PPQ_CONFIG.USING_CUDA_KERNEL:
        raise PermissionError("PPQ Floating Quant Function require PPQ_CONFIG.USING_CUDA_KERNEL = True")
    if not tensor.is_cuda:
        raise PermissionError(
            "PPQ Floating Quant Function requires tensor device to be cuda, "
            "CPU floating quantization is not implemented yet."
        )

    """PPQ 核心量化函数，没啥好说的了吧，这个玩意既做 quant 也做 dequant"""
    if not QuantizationStates.is_activated(config.state):
        return tensor
    if not config.policy.has_property(QuantizationProperty.FLOATING):
        raise ValueError(
            "Critical Quantization Error! Unexpected policy detected. "
            "PPQFloatingQuantFunction except a Floating Quantization Config."
        )
    if config.policy.has_property(QuantizationProperty.DYNAMIC):
        raise ValueError("Unexpected Dynamic Flag in Quantization Policy.")

    if config.policy.has_property(QuantizationProperty.PER_CHANNEL):
        return ChannelwiseFloatingQuantImpl.apply(
            tensor,
            config.scale,
            config.offset,
            config.channel_axis,
            config.exponent_bits,
            config.mantissa_bits,
            config.quant_min,
            config.quant_max,
            config.rounding,
        )
    elif config.policy.has_property(QuantizationProperty.PER_TENSOR):
        return TensorwiseFloatingQuantImpl.apply(
            tensor,
            config.scale,
            config.offset,
            config.exponent_bits,
            config.mantissa_bits,
            config.quant_min,
            config.quant_max,
            config.rounding,
        )



from ..core import NetworkFramework, TargetPlatform
from ..IR import BaseGraph, GraphBuilder, GraphExporter
from .native import NativeExporter, NativeImporter
from .onnx_exporter import OnnxExporter
from .onnx_parser import OnnxParser
from .onnxruntime_exporter import ONNXRUNTIMExporter



from . import ppq
from .ppq import (
    DataType,
    empty_ppq_cache,
    QuantizationVisibility,
    convert_any_to_torch_tensor,
    PPQ_CONFIG,
    TargetPlatform,
    TensorQuantizationConfig,
    OperationQuantizationConfig,
    QuantizationPolicy,
    QuantizationProperty,
    QuantizationStates,
    RoundingPolicy,
    common as ppq_common,
    ppq_quant_param_computing_function,
)
from .ppq.IR import BaseGraph, Operation, QuantableOperation, Variable, QuantableGraph, GraphReplacer
from .ppq.IR.search import SearchableGraph
from .ppq.IR.base.opdef import Opset
from .ppq.IR import GraphMerger, GraphFormatter
from .ppq.executor import TorchExecutor, BaseGraphExecutor
from .ppq.scheduler import DISPATCHER_TABLE, GraphDispatcher
from .ppq.IR.base.command import QuantizeOperationCommand
from .ppq.quantization.optim import (
    QuantizationOptimizationPass,
    QuantizationOptimizationPipeline,
    QuantizeFusionPass,
    QuantizeSimplifyPass,
    ParameterQuantizePass,
    ParameterBakingPass,
    LearnedStepSizePass,
    LayerwiseEqualizationPass,
)
from .ppq.quantization.algorithm.training import LSQDelegator, TrainableBlock
from .ppq.parser import OnnxParser
from .ppq.parser import ONNXRUNTIMExporter
from .ppq.quantization.measure import torch_mean_square_error, torch_snr_error, torch_cosine_similarity
from .ppq.quantization.qfunction.linear import PPQLinearQuantFunction, PPQLinearQuant_toInt
from .ppq.quantization import observer as ppq_observer
from .ppq.quantization.observer import BaseTensorObserver, TorchMinMaxObserver, OperationObserver
from .ppq.quantization.observer.range import minmax_to_scale_offset
from .ppq.quantization.measure import torch_KL_divergence


def load_onnx_graph(file: str):
    return OnnxParser().build(file)



from setuptools import find_packages, setup


def readme():
    with open("README.md", encoding="utf-8") as f:
        content = f.read()
    return content


def version():
    with open("VERSION_NUMBER", encoding="utf-8") as f:
        content = f.read().strip()
    return content


def license():
    with open("LICENSE", encoding="utf-8") as f:
        content = f.read()
    return content


setup(
    author="SpacemiT",
    author_email="xslim@spacemit.com",
    version=version(),
    description="XSlim is an offline quantization tools based on PPQ",
    long_description=readme(),
    long_description_content_type="text/markdown",
    install_requires=open("requirements.txt").readlines(),
    python_requires=">=3.9",
    name="xslim",
    packages=find_packages(),
    classifiers=[
        "Development Status :: 3 - Alpha",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.9",
    ],
    license=license(),
    include_package_data=True,
    zip_safe=False,
)



import json
from typing import Union

import numpy as np
import onnx
import torch
from onnx import helper, numpy_helper
from ..core import (
    GRAPH_OPSET_ATTRIB,
    ONNX_EXPORT_OPSET,
    ONNX_VERSION,
    PPQ_CONFIG,
    DataType,
    QuantizationStates,
    convert_any_to_numpy,
    ppq_warning,
)
from ..IR import BaseGraph, GraphExporter, Operation, OperationExporter, Variable
from ..IR.quantize import QuantableOperation


class ConstantOfShapeExporter(OperationExporter):
    def export(self, op: Operation, graph: BaseGraph, **kwargs) -> Operation:
        # PATCH 20211203, ConstantOfShape Op causes an export error.
        # 这一问题是由 ConstantOfShape 中的 value 格式问题引发的，下面的代码将导出正确的格式
        op.attributes["value"] = numpy_helper.from_array(op.attributes["value"])
        return op


class MMCVExporter(OperationExporter):
    def export(self, op: Operation, graph: BaseGraph, **kwargs) -> Operation:
        # MMCV operation must have a domain attribute.
        op.attributes["domain"] = "mmcv"
        return op


class InterpExporter(OperationExporter):
    def export(self, op: Operation, graph: BaseGraph, **kwargs) -> Operation:
        # PATCH 20211216, interp op can not export input_shape attribute.
        op.attributes.pop("input_shape")
        return op


class OOSExporter(OperationExporter):
    def export(self, op: Operation, graph: BaseGraph, **kwargs) -> Operation:
        # MMCV operation must have a domain attribute.
        op.attributes["domain"] = "com.microsoft"
        return op


class AttentionExporter(OperationExporter):
    def export(self, op: Operation, graph: BaseGraph, **kwargs) -> Operation:
        # MMCV operation must have a domain attribute.
        op.attributes["domain"] = "com.microsoft"
        return op


class PPQBiasFusedMatMulExporter(OperationExporter):
    def export(self, op: Operation, graph: BaseGraph, **kwargs) -> Operation:
        if op.num_of_input == 3:
            bias = op.inputs[-1]
        assert bias.is_parameter and bias.value is not None, "MatMul Format Error"

        bias_op = graph.create_operation(op_type="Add")
        op.type = "MatMul"
        graph.insert_op_after(bias_op, op)
        graph.create_variable(value=bias.value, is_parameter=True, dest_ops=[bias_op])
        graph.remove_variable(op.inputs[-1])


OP_CONVERTERS = {
    "ConstantOfShape": ConstantOfShapeExporter,
    "MMCVRoiAlign": MMCVExporter,
    "grid_sampler": MMCVExporter,
    "Interp": InterpExporter,
    "Attention": AttentionExporter,
    "QAttention": OOSExporter,
    "QGemm": OOSExporter,
    "QLinearAdd": OOSExporter,
    "QLinearAveragePool": OOSExporter,
    "QLinearConcat": OOSExporter,
    "QLinearConv": OOSExporter,
    "QLinearGlobalAveragePool": OOSExporter,
    "QLinearLeakyRelu": OOSExporter,
    "QLinearMul": OOSExporter,
    "QLinearReduceMean": OOSExporter,
    "QLinearSigmoid": OOSExporter,
    # 'PPQBiasFusedMatMul': PPQBiasFusedMatMulExporter
}


def convert_value(value: Union[int, float, np.ndarray, torch.Tensor]) -> str:
    if type(value) in {int, float}:
        return value
    else:
        value = convert_any_to_numpy(value, accept_none=True)
        if value is None:
            return value  # SOI config has Nona as its scale and
        return value.tolist()


class OnnxExporter(GraphExporter):
    """
    PPQ 可以将 计算图 导出成 Onnx 标准格式，Onnx Exporter 不会导出 QDQ 节点。
    如需导出带有 QDQ 节点的 Onnx 文件，用户需要使用 OnnxRuntime Exporter

    任何导出器的导出逻辑都是原地进行的，它们将对传入的计算图对象进行原地修改，因此在导出之前你需要手动克隆计算图。
    """

    def __init__(self) -> None:
        super().__init__()

    def export_quantization_config(self, config_path: str, graph: BaseGraph):
        """Export Tensor Quantization Config to File(Json)."""

        render_buffer = {"configs": {}, "dispatchings": {}, "values": {}}

        # Render quantization config.
        for operation in graph.operations.values():
            if isinstance(operation, QuantableOperation):
                op_dict = {
                    var.name: {
                        "bit_width": config.num_of_bits,
                        "policy": config.policy.to_dict(),
                        "state": config.state.name,
                        "quant_min": config.quant_min,
                        "quant_max": config.quant_max,
                        "hash": config.__hash__(),
                        "dominator": config.dominated_by.__hash__(),
                    }
                    for config, var in operation.config_with_variable
                }

                for config, _ in operation.config_with_variable:
                    if config.dominated_by == config:
                        if config.state != QuantizationStates.FP32:
                            render_buffer["values"][config.__hash__()] = {
                                "scale": convert_value(config.scale),
                                "zero_point": convert_value(config.offset),
                            }

                render_buffer["configs"][operation.name] = op_dict
                render_buffer["dispatchings"][operation.name] = operation.platform.name

        with open(file=config_path, mode="w") as file:
            json.dump(render_buffer, file, indent=4)

    def export_graph(self, graph: BaseGraph) -> onnx.ModelProto:
        """
        Convert a PPQ IR to Onnx IR.
        This export will only convert PPQ Op and var to onnx, all quantization configs will be skipped.

        This function will try to keep the opset version of your graph unchanged.
        However if the opset is not given, ppq will convert it to with the global parameter ppq.core.ONNX_EXPORT_OPSET.
        """
        name = graph._name
        if not name:
            name = f"{PPQ_CONFIG.NAME} - v({PPQ_CONFIG.VERSION})"

        # Ready to export onnx graph defination.
        _inputs, _outputs, _initilizers, _nodes, _value_info = [], [], [], [], []

        # before we can export them, we firstly convert all ops to proper format.
        for op in [_ for _ in graph.topological_sort()]:
            if op.type in OP_CONVERTERS:
                exporter = OP_CONVERTERS[op.type]()
                assert isinstance(
                    exporter, OperationExporter
                ), f"Expected an OpExporter here, however {type(exporter)} was given."
                op = exporter.export(op=op, graph=graph)

        for op in graph.topological_sort():
            _nodes.append(self.build_operator_proto(op))

        for variable in graph.variables.values():
            tensor_proto = self.build_variable_proto(variable)
            if variable.name in graph.inputs:
                _inputs.append(tensor_proto)
            if variable.name in graph.outputs:
                _outputs.append(tensor_proto)
            if variable.is_parameter:
                _initilizers.append(tensor_proto)
            else:
                _value_info.append(tensor_proto)

        graph_def = helper.make_graph(
            name=name, nodes=_nodes, inputs=_inputs, outputs=_outputs, initializer=_initilizers, value_info=_value_info
        )

        # if opset is missing from your graph, give it a default one.
        if GRAPH_OPSET_ATTRIB not in graph._detail:
            op = onnx.OperatorSetIdProto()
            op.version = ONNX_EXPORT_OPSET
            opsets = [op]
        else:
            opsets = []
            for opset in graph._detail[GRAPH_OPSET_ATTRIB]:
                op = onnx.OperatorSetIdProto()
                op.domain = opset["domain"]
                op.version = opset["version"]
                opsets.append(op)

        onnx_model = helper.make_model(graph_def, producer_name=PPQ_CONFIG.NAME, opset_imports=opsets)
        onnx_model.ir_version = graph._detail.get("ir_version", ONNX_VERSION)
        return onnx_model

    def build_operator_proto(self, operation: Operation) -> onnx.OperatorProto:
        """
        Convert PPQ Op to Onnx Operation
        An Op consumes zero or more Tensors, and produces zero or more Tensors.
        """
        attributes = operation.attributes
        for key in attributes:
            value = attributes[key]
            if isinstance(value, DataType):
                attributes[key] = value.value
            if isinstance(value, torch.Tensor):
                if value.numel() == 0:
                    attributes[key] = None
                elif value.numel() == 1:
                    attributes[key] = convert_any_to_numpy([value.item()])  # convert to 1d array
                else:
                    attributes[key] = convert_any_to_numpy(value)

        if PPQ_CONFIG.EXPORT_PPQ_INTERNAL_INFO:
            attributes["platform"] = operation.platform.name

        op_proto = helper.make_node(
            op_type=operation.type,
            inputs=[_.name for _ in operation.inputs],
            outputs=[_.name for _ in operation.outputs],
            name=operation.name,
            **attributes,
        )

        return op_proto

    def build_variable_proto(self, variable: Variable) -> onnx.TensorProto:
        """
        Convert PPQ Variable to Onnx TensorProto, There are 2 different types of Tensor in Onnx:
            Variable: Represents a Tensor whose value is not known until inference-time.
            Constant: Represents a Tensor whose value is known.
        """
        # Parameter Varaible in PPQ, Constant Variable in Onnx
        if variable.is_parameter:
            if variable.value is not None:
                var_shape = variable.value.shape
                pytorch_dtype = variable.value.dtype
                onnx_dtype = DataType.convert_from_torch(pytorch_dtype).value

        # Non Parameter
        else:
            var_shape = variable.shape
            onnx_dtype = variable.dtype.value

        if not variable.is_parameter:
            tensor_proto = helper.make_tensor_value_info(name=variable.name, elem_type=onnx_dtype, shape=var_shape)
        else:
            value = variable.value
            is_raw_format = False
            if isinstance(value, torch.Tensor):
                if value.numel() == 0:
                    value = []
                elif value.ndim >= 1:
                    value = convert_any_to_numpy(variable.value).flatten()
                    value = value.tobytes()
                    is_raw_format = True
                elif value.ndim == 0:  # Pytorch Scalar Type
                    value = [
                        value.item(),
                    ]  # it is fine for onnx, shape for this value will be []
            else:
                value = value  # value is python primary type.
            tensor_proto = helper.make_tensor(
                name=variable.name, data_type=onnx_dtype, dims=var_shape, vals=value, raw=is_raw_format
            )
        return tensor_proto

    def export(self, file_path: str, graph: BaseGraph, config_path: str = None, save_as_external_data: bool = False):
        # during export we will remove all boundary operations from graph.
        # we do not want to change the structure of original graph,
        # so there have to take a clone of it.
        # graph = graph.copy()

        # if a valid config path is given, export quantization config to there.
        if config_path is not None:
            self.export_quantization_config(config_path, graph)

        size_threshold = 0 if save_as_external_data else 1024
        onnx.save(
            self.export_graph(graph=graph),
            file_path,
            size_threshold=size_threshold,
            save_as_external_data=save_as_external_data,
            all_tensors_to_one_file=(not save_as_external_data),
        )



#!/usr/bin/env python3
# Copyright (c) 2023 SpacemiT. All rights reserved.
import functools
import os
from collections import OrderedDict
from enum import Enum
from typing import Callable, Dict, Iterable, List, Set, Tuple, Union

import numpy as np
import torch
from xslim.logger import logger

from ..defs import (PASSIVE_OPERATIONS, XQUANT_CONFIG, AutoFinetuneLevel,
                    PrecisionLevel)
from ..optimizer import (ActivationClipRefine, AsymmetricaUnsignlAlignSign,
                         ComputingFusionPass, FlattenGemmFusionPass,
                         FormatBatchNormalizationPass, HardSwishFusionPass,
                         PassiveParameterBakingPass, QuantizeConfigRefinePass,
                         RuntimeBlockWiseCalibrationPass, SwishFusionPass,
                         XSlimLayerwiseEqualizationPass)
from ..ppq_decorator import (BaseGraph, BaseGraphExecutor, GraphReplacer,
                             Operation, OperationQuantizationConfig,
                             ParameterBakingPass, ParameterQuantizePass,
                             QuantableGraph, QuantableOperation,
                             QuantizationOptimizationPipeline,
                             QuantizationPolicy, QuantizationProperty,
                             QuantizationStates, QuantizeFusionPass,
                             QuantizeOperationCommand, QuantizeSimplifyPass,
                             RoundingPolicy, TargetPlatform,
                             TensorQuantizationConfig, Variable, ppq_common)
from ..xslim_setting import XSlimSetting


def _get_quant_min_max(num_of_bits: int, signed: bool = True):
    if signed:
        return -(2 ** (num_of_bits - 1)), 2 ** (num_of_bits - 1) - 1
    else:
        return 0, 2**num_of_bits - 1


class XSlimQuantizer:
    def __init__(self, graph: BaseGraph) -> Union[torch.Tensor, list, dict]:
        self._graph = graph
        self._processor = QuantableGraph(GraphReplacer(self._graph))
        self._precision_level = PrecisionLevel.LEVEL_0
        self._num_of_bits = 8
        self._quant_min, self._quant_max = _get_quant_min_max(self._num_of_bits)
        perchannel_policy = QuantizationPolicy(
            QuantizationProperty.SYMMETRICAL + QuantizationProperty.PER_CHANNEL + QuantizationProperty.LINEAR
        )
        pertensor_policy = QuantizationPolicy(
            QuantizationProperty.SYMMETRICAL
            + QuantizationProperty.PER_TENSOR
            + QuantizationProperty.LINEAR
            + QuantizationProperty.POWER_OF_2
        )
        self._op_type_to_policy = {
            "Conv": perchannel_policy,
            "ConvTranspose": pertensor_policy,
            "Gemm": pertensor_policy,
            "MatMul": pertensor_policy,
            "BatchMatMul": pertensor_policy,
        }
        self._observer_mapping = {
            "default": "xslim",
            "kl": "kl",
            "mse": "mse",
            "minmax": "minmax",
            "percentile": "percentile",
        }
        self._verbose = False
        self._bias_to_fp32 = bool(int(os.environ.get("XQUANT_BIAS_TO_FP32", "1")) > 0)

        if self._bias_to_fp32:
            logger.info("Set Bias to Float32.")

    @property
    def target_platform(self) -> TargetPlatform:
        return TargetPlatform.XQUANT_INT8

    def quantize_operation(self, op_name: str, platform: TargetPlatform = None) -> QuantableOperation:
        if op_name not in self._graph.operations:
            raise KeyError(f"Can not find op {op_name} in your graph, chech operation name again.")
        converting_operation = self._graph.operations[op_name]
        if isinstance(converting_operation, QuantableOperation):
            logger.warning(f"Operation {op_name} has been quantized, can not to quantize it twice.")
            return converting_operation

        # override platform with calling parameter.
        if platform is not None:
            converting_operation.platform = platform
        else:
            platform = converting_operation.platform

        if platform in {TargetPlatform.FP32, TargetPlatform.SOI}:
            return self._graph.operations[op_name]

        # if platform == TargetPlatform.UNSPECIFIED we can skip its quantization when type is not supported.
        if platform == TargetPlatform.UNSPECIFIED and converting_operation.type not in self.quant_operation_types:
            return self._graph.operations[op_name]

        # create quantize config and convert operation.
        self._processor(
            QuantizeOperationCommand(
                op_name=op_name,
                target_platform=platform,
                config=self.init_quantize_config(operation=converting_operation),
            )
        )
        return self._graph.operations[op_name]

    def quantize(
        self,
        inputs: Union[torch.Tensor, list, dict],
        calib_dataloader: Iterable,
        executor: BaseGraphExecutor,
        xslim_setting: XSlimSetting,
        **kwargs,
    ) -> None:
        self._set_max_percentile = xslim_setting.quantization_parameters.max_percentile
        self._calibration_type = xslim_setting.calibration_parameters.calibration_type

        executor.load_graph(self._graph)
        executor.tracing_operation_meta(inputs=inputs)
        # step - 1, prequant pipeline:
        # prequant pipeline will change your network structure and float value.
        prequant_pipeline = self.build_prequant_pipeline(xslim_setting, executor=executor)
        prequant_pipeline.optimize(
            graph=self._graph, dataloader=calib_dataloader, executor=executor, verbose=self._verbose, **kwargs
        )

        # step - 2, quantize all operations
        executor.load_graph(self._graph)
        executor.tracing_operation_meta(inputs=inputs)

        for op_name, operation in self._graph.operations.items():
            if operation.platform == TargetPlatform.UNSPECIFIED:
                if operation.type in self.quant_operation_types:
                    operation.platform = self.target_platform
                else:
                    operation.platform = TargetPlatform.FP32

            if operation.platform not in {TargetPlatform.FP32, TargetPlatform.SOI}:
                self.quantize_operation(op_name)

        # quantize operation will modify network structure
        # it is necessary calling self._executor before further execution
        # step - 3, calling graph optimization pipeline
        executor.load_graph(self._graph)
        quant_pipeline = self.build_quant_pipeline(xslim_setting)

        quant_pipeline.optimize(
            graph=self._graph, dataloader=calib_dataloader, executor=executor, verbose=self._verbose, **kwargs
        )

    @staticmethod
    def create_default_quant_config(
        op: Operation,
        num_of_bits: int = 8,
        quant_min: Union[int, float] = -127,
        quant_max: Union[int, float] = 128,
        observer_algorithm: str = "percentile",
        policy: QuantizationPolicy = QuantizationPolicy(
            QuantizationProperty.PER_TENSOR + QuantizationProperty.LINEAR + QuantizationProperty.SYMMETRICAL
        ),
        rounding: RoundingPolicy = RoundingPolicy.ROUND_HALF_EVEN,
        exponent_bits: int = 0,
    ) -> OperationQuantizationConfig:
        assert isinstance(op, Operation), f"Can only initialize OQC for PPQ.IR.Operation, however {type(op)} was given."
        assert isinstance(
            policy, QuantizationPolicy
        ), f"Can not create quantization config - Quantization Policy Type Error."
        assert isinstance(rounding, RoundingPolicy), f"Can not create quantization config - Rounding Policy Type Error."

        socket = op.socket
        input_cfgs, output_cfgs = [], []
        for index in range(op.num_of_input):
            state = QuantizationStates.INITIAL
            # for those unexpected inputs and outputs
            # ppq just initilize them as normal variable.
            if index < len(socket.in_plat):
                target_plat = socket.in_plat[index]
                if target_plat == TargetPlatform.FP32:
                    state = QuantizationStates.FP32
                if target_plat == TargetPlatform.SOI:
                    state = QuantizationStates.FP32
            input_cfgs.append(
                TensorQuantizationConfig(
                    policy=policy,
                    rounding=rounding,
                    num_of_bits=num_of_bits,
                    scale=None,
                    offset=None,
                    exponent_bits=exponent_bits,
                    quant_min=quant_min,
                    quant_max=quant_max,
                    observer_algorithm=observer_algorithm,
                    state=state,
                )
            )

        for index in range(op.num_of_output):
            state = QuantizationStates.INITIAL
            # for those unexpected inputs and outputs
            # ppq just initilize them as normal variable.
            if index < len(socket.out_plat):
                target_plat = socket.out_plat[index]
                if target_plat == TargetPlatform.FP32:
                    state = QuantizationStates.FP32
                if target_plat == TargetPlatform.SOI:
                    state = QuantizationStates.FP32
            output_cfgs.append(
                TensorQuantizationConfig(
                    policy=policy,
                    rounding=rounding,
                    num_of_bits=num_of_bits,
                    scale=None,
                    offset=None,
                    exponent_bits=exponent_bits,
                    quant_min=quant_min,
                    quant_max=quant_max,
                    observer_algorithm=observer_algorithm,
                    state=state,
                )
            )

        return OperationQuantizationConfig(input_cfgs, output_cfgs)

    def report(self):
        pass

    def init_quantize_config(self, operation: Operation) -> OperationQuantizationConfig:
        observer = "default" if self._calibration_type is None else self._calibration_type
        base_quant_config = self.create_default_quant_config(
            policy=self.quantize_policy,
            rounding=self.rounding_policy,
            op=operation,
            num_of_bits=self._num_of_bits,
            exponent_bits=0,
            quant_max=self._quant_max,
            quant_min=self._quant_min,
            observer_algorithm=self._observer_mapping.get(observer),
        )
        # 对常量输入永远采用minmax
        for in_var, in_tqc in zip(operation.inputs, base_quant_config.input_quantization_config):
            if in_var.is_parameter:
                in_tqc.observer_algorithm = "minmax"

        if isinstance(self._set_max_percentile, float):
            for in_var, in_tqc in zip(operation.inputs, base_quant_config.input_quantization_config):
                if not in_var.is_parameter:
                    in_tqc.detail[ppq_common.OBSERVER_PERCENTILE_MANUL_OVERRIDE] = self._set_max_percentile
            for out_var, out_tqc in zip(operation.outputs, base_quant_config.output_quantization_config):
                if not out_var.is_parameter:
                    out_tqc.detail[ppq_common.OBSERVER_PERCENTILE_MANUL_OVERRIDE] = self._set_max_percentile

        if operation.type in {"Conv", "ConvTranspose", "Gemm", "MatMul"}:
            # set all parameters within Conv, ConvTranspose, Gemm to per-channel quant-config.
            assert operation.num_of_input > 0, "Seems you got a Conv layer with no parameters."
            # set weight tqc
            if operation.inputs[1].is_parameter:
                in_tqc = base_quant_config.input_quantization_config[1]
                in_tqc.num_of_bits = 8
                in_tqc._quant_min, in_tqc._quant_max = _get_quant_min_max(in_tqc.num_of_bits)
                in_tqc.policy = self._op_type_to_policy[operation.type]
                in_tqc.observer_algorithm = "minmax"
                if in_tqc.policy.has_property(QuantizationProperty.PER_CHANNEL):
                    in_tqc.channel_axis = 1 if operation.type == "ConvTranspose" else 0
                    if operation.type == "Gemm" and operation.attributes.get("transB", 0) == 0:
                        in_tqc.channel_axis = 1

            # if operation has bias
            if operation.num_of_input > 2 and operation.inputs[-1].is_parameter:
                in_tqc = base_quant_config.input_quantization_config[-1]
                if self._bias_to_fp32:
                    in_tqc.state = QuantizationStates.FP32
                else:
                    in_tqc.policy = self._op_type_to_policy[operation.type]
                    in_tqc.num_of_bits = 32
                    in_tqc._quant_min, in_tqc._quant_max = _get_quant_min_max(in_tqc.num_of_bits)
                    in_tqc.state = QuantizationStates.PASSIVE_INIT
                    if in_tqc.policy.has_property(QuantizationProperty.PER_CHANNEL):
                        in_tqc.channel_axis = 0

        elif operation.type in {"BatchMatMul"}:
            for in_tqc in base_quant_config.input_quantization_config[2:]:
                in_tqc.state = QuantizationStates.FP32
            if operation.inputs[1].is_parameter:
                in_tqc = base_quant_config.input_quantization_config[1]
                in_tqc.num_of_bits = 8
                in_tqc._quant_min, in_tqc._quant_max = _get_quant_min_max(in_tqc.num_of_bits)
                in_tqc.policy = self._op_type_to_policy[operation.type]
                in_tqc.observer_algorithm = "minmax"

        elif operation.type in {"LayerNormalization", "LayerNorm", "InstanceNormalization", "BatchNormalization"}:
            for in_tqc in base_quant_config.input_quantization_config[1:]:
                in_tqc.state = QuantizationStates.FP32

        return base_quant_config

    @property
    def quant_operation_types(self) -> set:
        QUANTTYPE = {
            "Conv",
            "ConvTranspose",
            "Gemm",
            "MatMul",
            "BatchMatMul",
            #
            "Relu",
            "PRelu",
            "LeakyRelu",
            "Sigmoid",
            "HardSwish",
            "HardSigmoid",
            "Gelu",
            "GELU",
            "LRN",
            "Clip",
            #
            "Pad",
            #
            "Resize",
            #
            "MaxPool",
            "AveragePool",
            "GlobalMaxPool",
            "GlobalAveragePool",
            "ReduceMean",
            #
            "Add",
            "Sub",
            "Mul",
            # "Div",
            # "Max",
            #
            "LayerNormalization",
            "LayerNorm",
            "InstanceNormalization",
            "GroupNormalization",
            #
            "Gather",
            "Reshape",
            "Concat",
            "Split",
            "Transpose",
            "Slice",
            "Flatten",
        }
        QUANTTYPE.update(PASSIVE_OPERATIONS)
        return QUANTTYPE

    @property
    def quantize_policy(self) -> QuantizationPolicy:
        return QuantizationPolicy(
            QuantizationProperty.ASYMMETRICAL + QuantizationProperty.PER_TENSOR + QuantizationProperty.LINEAR
        )

    @property
    def rounding_policy(self) -> RoundingPolicy:
        return RoundingPolicy.ROUND_HALF_EVEN

    @property
    def activation_fusion_types(self) -> set:
        return {"Relu", "Clip"}

    def build_quant_pipeline(self, setting: XSlimSetting) -> QuantizationOptimizationPipeline:
        list_of_passes = []

        list_of_passes.append(
            QuantizeConfigRefinePass(
                setting.quantization_parameters.precision_level.value,
                setting.quantization_parameters.custom_setting,
            )
        )

        list_of_passes.append(
            QuantizeFusionPass(
                fuse_activation=True,
                fuse_passive_op=True,
                fuse_relu_clip=True,
                activation_type=self.activation_fusion_types,
            )
        )
        list_of_passes.append(QuantizeSimplifyPass())

        list_of_passes.append(ActivationClipRefine())

        list_of_passes.append(PassiveParameterBakingPass())

        list_of_passes.append(ParameterQuantizePass(method="minmax"))

        list_of_passes.append(
            RuntimeBlockWiseCalibrationPass(
                self._observer_mapping.get(setting.calibration_parameters.calibration_type, "xslim"),
                block_wise=True,
                fintune_epoch=XQUANT_CONFIG.fine_tune_epoch,
                auto_finetune_level=setting.quantization_parameters.finetune_level.value,
            )
        )

        list_of_passes.append(ParameterBakingPass())
        list_of_passes.append(PassiveParameterBakingPass())
        return QuantizationOptimizationPipeline(passes=list_of_passes)

    def build_prequant_pipeline(
        self, setting: XSlimSetting, executor: BaseGraphExecutor
    ) -> QuantizationOptimizationPipeline:
        list_of_passes = []

        finetune_level = setting.quantization_parameters.finetune_level.value

        list_of_passes.append(FlattenGemmFusionPass())
        list_of_passes.append(FormatBatchNormalizationPass())

        if finetune_level >= AutoFinetuneLevel.DO_NOTHING.value:
            list_of_passes.append(
                XSlimLayerwiseEqualizationPass(
                    optimize_level=2 if finetune_level > AutoFinetuneLevel.DO_NOTHING.value else 1,
                    including_act=finetune_level > AutoFinetuneLevel.DO_NOTHING.value,
                )
            )

        return QuantizationOptimizationPipeline(passes=list_of_passes)



from typing import Union

import numpy as np
import torch
from ..core import DataType, convert_any_to_numpy


def convert_value(
    value: Union[int, float, np.ndarray, torch.Tensor], export_as_float: bool, dtype: DataType = DataType.FP32
) -> Union[float, list]:
    """Converting value from any to python native data dtype, ready for export.

    Args:
        value (Union[int, float, np.ndarray, torch.Tensor]): exporting value.
        export_as_list (bool): export value as a list.
        dtype (DataType, optional): exporting dtype.

    Returns:
        Union[float, list]: Converted value
    """
    if dtype not in {DataType.FP32, DataType.INT32}:
        raise ValueError(
            f"Can Only export dtype fp32 and int32, " f"while you are requiring to dump a {dtype.name} value"
        )
    value = convert_any_to_numpy(value, accept_none=False)
    value = value.astype(dtype=DataType.to_numpy(dtype))
    if export_as_float:
        value = value.item()
        assert type(value) in {int, float}, (
            f"Trying to dump a tensorwise quantization value {value}. "
            f"It is Expected to be a int or float value, while {type(value)} was given"
        )
        return value
    else:
        value = convert_any_to_numpy(value, accept_none=False)
        return value.tolist()



#!/usr/bin/env python3
# Copyright (c) 2023 SpacemiT. All rights reserved.
import functools
import random
from collections import defaultdict
from typing import Callable, Dict, Iterable, List, Sequence, Set, Tuple, Union

import numpy as np
import torch
from tqdm import tqdm
from xslim.logger import logger

from ..defs import XQUANT_CONFIG
from ..ppq_decorator import (
    BaseGraph,
    BaseGraphExecutor,
    LearnedStepSizePass,
    LSQDelegator,
    Operation,
    PPQLinearQuantFunction,
    QuantableGraph,
    QuantableOperation,
    QuantizationProperty,
    QuantizationStates,
    TensorQuantizationConfig,
    TorchExecutor,
    TrainableBlock,
    Variable,
    torch_mean_square_error,
    torch_snr_error,
)


class XSlimTrainableBlock(TrainableBlock):
    OP_DEPTH_DICT = {
        "Conv": 4,
        "ConvTranspose": 4,
        "Gemm": 4,
        "MatMul": 4,
        "LayerNormalization": 2,
        "InstanceNormalization": 2,
        "GroupNormalization": 2,
        "MaxPool": 2,
        "AveragePool": 2,
        "GlobalMaxPool": 2,
        "GlobalAveragePool": 2,
        "ReduceMean": 2,
        "Reshape": 0.5,
        "Transpose": 0.5,
        "Slice": 0.5,
        "Squeeze": 0.5,
        "Unsqueeze": 0.5,
        "Gather": 0.5,
        "Flatten": 0.5,
        "Split": 0.5,
    }

    def __init__(self, sp: Operation, ep: Operation, rps: List[Operation], graph: BaseGraph):
        super().__init__(sp, ep, rps)
        self.graph: BaseGraph = graph
        self.in_var_names = set()
        self.out_var_names = set()
        self.depth = 0
        self.update_block_io_var_names()
        self.get_depth()

    @staticmethod
    def get_sequence_block_depth(rps: List[Operation]):
        depth = 0
        for op in rps:
            depth += XSlimTrainableBlock.OP_DEPTH_DICT.get(op.type, 1)
        return depth

    def get_depth(self):
        block_ops = set(self.rps)
        toposort_ops = []
        visited_vars = set()

        def visit_ops(in_vars: List[Variable], depth):
            next_vars = []
            max_depth = 0
            for var in in_vars:
                for dest_op in var.dest_ops:
                    if (
                        dest_op in block_ops
                        and set([var for var in dest_op.inputs if not var.is_parameter and var.source_op is not None])
                        <= visited_vars
                    ):
                        toposort_ops.append(dest_op)
                        block_ops.remove(dest_op)
                        next_vars.extend(dest_op.outputs)
                        visited_vars.update([var for var in dest_op.outputs])
                        max_depth = max(XSlimTrainableBlock.OP_DEPTH_DICT.get(dest_op.type, 1), max_depth)
            depth = max_depth + depth
            if len(next_vars) > 0:
                return visit_ops(next_vars, depth)
            return depth

        in_vars = [self.graph.variables[name] for name in self.in_var_names]
        visited_vars.update(in_vars)
        depth = visit_ops(in_vars, 0)
        if len(block_ops) > 0:
            raise RuntimeError(
                "get block depth and toposort error, {} no visited".format([op.name for op in block_ops])
            )
        self.depth = depth
        self.rps = toposort_ops
        return depth

    def update_block_io_var_names(self) -> Tuple[Set[str], Set[str]]:
        block_op_set = set(self.rps)
        block_output_names_set = set()
        block_input_names_set = set()
        for operation in self.rps:
            for o_var in operation.outputs:
                if isinstance(o_var.dest_ops, Sequence) and len(o_var.dest_ops) > 0:
                    for dest_op in o_var.dest_ops:
                        if dest_op not in block_op_set:
                            block_output_names_set.add(o_var.name)
                            break
                else:
                    block_output_names_set.add(o_var.name)
            for i_var in operation.inputs:
                if i_var.source_op is not None:
                    if i_var.source_op not in block_op_set:
                        block_input_names_set.add(i_var.name)
                elif not i_var.is_parameter and i_var.name != "":
                    block_input_names_set.add(i_var.name)
        self.in_var_names = block_input_names_set
        self.out_var_names = block_output_names_set
        return block_input_names_set, block_output_names_set


class XSlimBlockBuilder:
    def __init__(self, graph: BaseGraph, topo_order: List[Operation]) -> None:
        self.graph = graph
        self.op_orders = topo_order
        self.sequence_blocks: List[XSlimTrainableBlock] = []
        self.var_to_dest_blocks = dict()
        self.var_to_src_blocks = dict()
        self.sequence_init()
        self.sequence_io_init()

    def sequence_init(self):
        """

        初始化一个单序列的块列表
        """
        visited_ops = set()

        def _find_coherent_ops(s_op: Operation, rps: List[Operation]) -> Operation:
            rps.append(s_op)
            downstrem_ops = self.graph.get_downstream_operations(s_op)
            upstrem_ops = self.graph.get_upstream_operations(s_op)
            if (
                len(downstrem_ops) == 1
                and len(upstrem_ops) <= 1
                and len(self.graph.get_upstream_operations(downstrem_ops[0])) == 1
                and XSlimTrainableBlock.get_sequence_block_depth(rps) < XQUANT_CONFIG.min_block_size
            ):
                return _find_coherent_ops(downstrem_ops[0], rps)
            else:
                return s_op

        for op in self.op_orders:
            if op in visited_ops:
                continue
            s_op = op
            rps = []
            e_op = _find_coherent_ops(s_op, rps)
            for _op in rps:
                visited_ops.add(_op)
            self.sequence_blocks.append(XSlimTrainableBlock(s_op, e_op, rps, self.graph))

    def sequence_io_init(self):
        self.var_to_dest_blocks.clear()
        self.var_to_src_blocks.clear()
        for block in self.sequence_blocks:
            for var_name in block.in_var_names:
                if var_name not in self.var_to_dest_blocks:
                    self.var_to_dest_blocks[var_name] = []
                self.var_to_dest_blocks[var_name].append(block)

            for var_name in block.out_var_names:
                if var_name not in self.var_to_src_blocks:
                    self.var_to_src_blocks[var_name] = []
                self.var_to_src_blocks[var_name].append(block)

    def get_downstream_blocks(self, block: XSlimTrainableBlock):
        downstream_blocks = set()
        for var_name in block.out_var_names:
            downstream_blocks.update(self.var_to_dest_blocks.get(var_name, []))
        return list(downstream_blocks)

    def get_upstream_blocks(self, block: XSlimTrainableBlock):
        upstream_blocks = set()
        for var_name in block.in_var_names:
            upstream_blocks.update(self.var_to_src_blocks.get(var_name, []))
        return list(upstream_blocks)

    def build(self):
        def _find_same_input_block(block: XSlimTrainableBlock):
            visited_var_names = block.in_var_names | block.out_var_names
            merge_blocks = set()
            merge_block_list = []
            down_blocks = self.get_downstream_blocks(block)

            for _ in range(len(down_blocks)):
                # 因为一个块的合并，则另一个块也可能可以合并，但这个顺序不一定
                for down_block in down_blocks:
                    if down_block.in_var_names <= visited_var_names and down_block not in merge_blocks:
                        merge_blocks.add(down_block)
                        merge_block_list.append(down_block)
                        visited_var_names.update(down_block.out_var_names)
            return merge_block_list

        def _merge_block(
            start_block: XSlimTrainableBlock, end_block: XSlimTrainableBlock, rp_blocks: List[XSlimTrainableBlock]
        ):
            rp_blocks.append(start_block)
            if end_block is start_block:
                return
            for down_block in self.get_downstream_blocks(start_block):
                if down_block not in rp_blocks:
                    _merge_block(down_block, end_block, rp_blocks)

        for merge_step in range(XQUANT_CONFIG.merge_block_step):
            valid_blocks = []
            visited_block = set()
            for block in self.sequence_blocks:
                if block in visited_block:
                    continue
                visited_block.add(block)
                merge_blocks = []
                if block.depth < XQUANT_CONFIG.max_block_size:
                    merge_blocks = _find_same_input_block(block)

                update_block = block
                if len(merge_blocks) > 0:
                    max_add_depth = 0
                    for rp_block in merge_blocks:
                        max_add_depth = max(rp_block.depth, max_add_depth)

                    if (max_add_depth + block.depth) < XQUANT_CONFIG.max_block_size:
                        for rp_block in merge_blocks:
                            visited_block.add(rp_block)
                            update_block.rps.extend(rp_block.rps)
                        update_block.ep = merge_blocks[-1].ep
                        update_block.update_block_io_var_names()
                        update_block.get_depth()
                valid_blocks.append(update_block)

            self.sequence_blocks = valid_blocks
            self.sequence_io_init()

        for block in self.sequence_blocks:
            block.get_depth()

        return self.sequence_blocks


class LSQDelegatorDecorator(LSQDelegator):
    def finalize(self) -> None:
        with torch.no_grad():
            if isinstance(self.config.scale, torch.Tensor):
                if torch.any(self.config.scale < 0):
                    self.withdraw()

        self.scale_backup = None
        self.offset_backup = None
        self.param_backup = None

    def __call__(self, tensor: torch.Tensor, config: TensorQuantizationConfig) -> torch.Tensor:
        scale, offset = config.scale, config.offset

        def grad_scale(t, factor):
            return (t - (t * factor)).detach() + (t * factor)

        grad_factor = 1 / (tensor.numel() * config.quant_max) ** 0.5
        offset = (offset.round() - offset).detach() + offset
        if self.is_scale_trainable:
            scale = scale.abs()
            scale = grad_scale(scale, grad_factor)

        if self.is_offset_trainable:
            offset = grad_scale(offset, grad_factor)

        if config.policy.has_property(QuantizationProperty.PER_CHANNEL):
            shape = [1 if axis != config.channel_axis else -1 for axis in range(tensor.ndim)]
            scale = scale.view(shape)
            offset = offset.view(shape)

        quantized = tensor / scale + offset
        quantized = (quantized.round() - quantized).detach() + quantized
        quantized = torch.clamp(quantized, config.quant_min, config.quant_max)
        return (quantized - offset) * scale


class LearnedStepSizePassDecorator(LearnedStepSizePass):
    def collect(
        self,
        graph: BaseGraph,
        block: XSlimTrainableBlock,
        executor: TorchExecutor,
        dataloader: Iterable,
        collate_fn: Callable,
        collecting_device: str,
        steps: int = None,
        expire_device: str = "cpu",
    ) -> Tuple[List[Dict[str, torch.Tensor]], List[Dict[str, torch.Tensor]]]:
        def cache_fn(data: torch.Tensor):
            if collecting_device == "cuda":
                data = data.cuda()
                mem_free, mem_all = torch.cuda.mem_get_info()
                mem_free_ratio = mem_free / mem_all
                if mem_free_ratio < 0.2:
                    data = data.cpu()
            else:
                data = data.cpu()
            return data

        block_output_names = list(block.out_var_names)
        block_input_names = list(block.in_var_names)
        with torch.no_grad():
            quant_graph = QuantableGraph(graph)  # helper class
            fp_outputs, qt_inputs = [], []

            cur_iter = 0
            # dequantize graph, collect fp32 outputs
            quant_graph.dequantize_graph(expire_device=expire_device)
            for data in dataloader:
                if collate_fn is not None:
                    data = collate_fn(data)

                fp_output = executor.forward(data, block_output_names)
                fp_output = {var_name: cache_fn(value) for var_name, value in zip(block_output_names, fp_output)}
                fp_outputs.append(fp_output)
                cur_iter += 1
                if steps is not None and cur_iter > steps:
                    break

            cur_iter = 0
            # restore quantization state, collect quant inputs
            quant_graph.restore_quantize_state(expire_device=expire_device)
            for data in dataloader:
                if collate_fn is not None:
                    data = collate_fn(data)
                qt_input = executor.forward(data, block_input_names)
                qt_input = {var_name: cache_fn(value) for var_name, value in zip(block_input_names, qt_input)}
                qt_inputs.append(qt_input)
                cur_iter += 1
                if steps is not None and cur_iter > steps:
                    break

        return qt_inputs, fp_outputs

    def compute_block_loss(
        self,
        block: XSlimTrainableBlock,
        qt_inputs: List[Dict[str, torch.Tensor]],
        fp_outputs: List[Dict[str, torch.Tensor]],
        executor: TorchExecutor,
        loss_fn: Callable = torch_mean_square_error,
    ) -> float:
        with torch.no_grad():
            block_output_names = list(block.out_var_names)
            block_input_names = list(block.in_var_names)
            losses = {var_name: 0.0 for var_name in block_output_names}
            for qt_input, fp_output in zip(qt_inputs, fp_outputs):
                feed_dict = {k: v.to(executor._device) for k, v in qt_input.items()}

                qt_output = executor.partial_graph_forward(
                    operations=block.rps, feed_dict=feed_dict, output_names=block_output_names
                )

                for name, quant_output in zip(block_output_names, qt_output):
                    batch_loss = loss_fn(quant_output.to(executor._device), fp_output[name].to(executor._device))
                    losses[name] += batch_loss.detach().item()

            for name in losses:
                losses[name] /= len(qt_inputs)
        return sum([v for v in losses.values()])

    def finetune(
        self,
        steps: int,
        learning_rate: float,
        block: XSlimTrainableBlock,
        executor: TorchExecutor,
        qt_inputs: List[Dict[str, torch.Tensor]],
        fp_outputs: List[Dict[str, torch.Tensor]],
        optimizer: torch.optim.Optimizer = None,
    ) -> Tuple[float, float]:
        # step - 1: enable gradient for training.
        self.enable_block_gradient(block)

        # record pre training loss.
        pre_loss = self.compute_block_loss(
            block=block, qt_inputs=qt_inputs, fp_outputs=fp_outputs, executor=executor, loss_fn=self.loss_fn
        )

        # collect trainable params
        delegators = {}
        trainable_scales = []
        for op in block.rps:
            if not isinstance(op, QuantableOperation):
                continue

            # register quant delegator
            for cfg, var in op.config_with_variable:
                if cfg.detail.get("NONE_VALUE", False):
                    continue
                if cfg.state in {QuantizationStates.ACTIVATED}:
                    offset_trainable = (
                        cfg.policy.has_property(QuantizationProperty.ASYMMETRICAL) and self.is_scale_trainable
                    )
                    delegator = LSQDelegatorDecorator(
                        config=cfg,
                        var=var,
                        is_scale_trainable=self.is_scale_trainable,
                        is_offset_trainable=offset_trainable,
                    )
                    trainable_scales.extend(delegator.trainable_tensors())
                    executor.register_quantize_delegate(config=cfg, delegator=delegator)
                    delegators[cfg] = delegator
                elif cfg.state in {QuantizationStates.PASSIVE, QuantizationStates.PASSIVE_INIT} and var.is_parameter:
                    delegator = LSQDelegatorDecorator(
                        config=cfg,
                        var=var,
                        is_scale_trainable=False,
                        is_offset_trainable=False,
                    )
                    trainable_scales.extend(delegator.trainable_tensors())
                    executor.register_quantize_delegate(config=cfg, delegator=delegator)
                    delegators[cfg] = delegator

        # check if empty.
        tensors = [tensor for tensor in trainable_scales if tensor.requires_grad]
        tensors = set(tensors)  # remove duplicated tensor
        if len(tensors) == 0:
            for cfg, delegator in delegators.items():
                executor.remove_quantize_delegate(config=cfg)
            return 0, 0

        # initilize optimizer.
        if self.optimizer is None:
            optimizer = torch.optim.Adam(tensors, lr=learning_rate)
        else:
            optimizer = self.optimizer(tensors, lr=learning_rate)

        dataset_length = len(qt_inputs)
        if dataset_length == 0:
            raise ValueError("Dataset is empty.")

        range_steps = [i for i in range(steps)]
        random.shuffle(range_steps)
        for idx in tqdm(range_steps, desc="Block Tuning"):
            qt_input, fp_output = qt_inputs[idx % dataset_length], fp_outputs[idx % dataset_length]

            # forward
            optimizer.zero_grad()
            feed_dict = {k: v.to(executor._device) for k, v in qt_input.items()}
            output_names = [name for name in fp_output]

            qt_output = executor.partial_graph_forward(
                operations=block.rps, feed_dict=feed_dict, output_names=output_names
            )

            # compute loss
            loss = 0.0
            for name, o_var in zip(output_names, qt_output):
                loss += self.loss_fn(o_var, fp_output[name].to(executor._device))

            for op in block.rps:
                if self.gamma == 0:
                    continue
                if isinstance(op, QuantableOperation) and op.is_computing_op:
                    weight = op.inputs[1].value
                    wconfig = op.config.input_quantization_config[1]
                    loss += torch_mean_square_error(weight, PPQLinearQuantFunction(weight, wconfig)) * self.gamma

            # backward from loss
            assert isinstance(loss, torch.Tensor)
            loss.backward()
            optimizer.step()

        # step - 3: record post training loss
        post_loss = self.compute_block_loss(
            block=block, qt_inputs=qt_inputs, fp_outputs=fp_outputs, executor=executor, loss_fn=self.loss_fn
        )
        # check and withdraw
        if post_loss > pre_loss:
            for cfg, delegator in delegators.items():
                delegator.withdraw()
            logger.info(f"Tuning Finished: loss no change and withdraw.\n")
        else:
            logger.info(f"Tuning Finished: ({pre_loss:.5f} -> {min(pre_loss, post_loss):.5f})\n")

        for cfg, delegator in delegators.items():
            delegator.finalize()
            executor.remove_quantize_delegate(config=cfg)

        # disable gradient for evaluation.
        self.disable_block_gradient(block)

        # clear cache
        torch.cuda.empty_cache()

        return pre_loss, post_loss



from .base import (GraphDispatcher, reverse_tracing_pattern,
                   value_tracing_pattern)
from .dispatchers import AggresiveDispatcher, ConservativeDispatcher, PPLNNDispatcher, PointDispatcher
from .allin import AllinDispatcher
from .perseus import Perseus
# Do not forget register your dispather here.

DISPATCHER_TABLE = {
    "conservative": ConservativeDispatcher,
    "pplnn": PPLNNDispatcher,
    "aggresive": AggresiveDispatcher,
    "pointwise": PointDispatcher,
    "allin": AllinDispatcher,
    'perseus': Perseus
}


#!/usr/bin/env python3
# Copyright (c) 2023 SpacemiT. All rights reserved.
from .calibration import RuntimeBlockWiseCalibrationPass
from .equalization import XSlimLayerwiseEqualizationPass
from .fusion import (
    ComputingFusionPass,
    FlattenGemmFusionPass,
    FormatBatchNormalizationPass,
    HardSwishFusionPass,
    SwishFusionPass,
)
from .legalized import GraphLegalized
from .observer import TorchXSlimKLObserver, TorchXSlimMSEObserver, TorchXSlimObserver
from .refine import (
    ActivationClipRefine,
    AsymmetricaUnsignlAlignSign,
    PassiveParameterBakingPass,
    QuantizeConfigRefinePass,
    QuantizeFusionPass,
)
from .training import LearnedStepSizePassDecorator, LSQDelegatorDecorator



"""Regression tests for dynamic ONNX quantization helpers."""

import os
import sys
import unittest

import numpy as np
import onnx
from onnx import TensorProto, helper, numpy_helper
from onnxruntime.quantization.quant_utils import quantize_data as ort_quantize_data

sys.path.insert(0, os.path.join(os.path.dirname(__file__), ".."))

from xslim.quantizer.dynamic_quantizer import dynamic_quantize_onnx_model, dynamic_weight_only_quantize


class TestDynamicQuantizer(unittest.TestCase):
    """Validate dynamic weight-only quantization edge cases."""

    @staticmethod
    def _build_matmul_model(weight_value, opset_imports=None):
        x = helper.make_tensor_value_info("x", TensorProto.FLOAT, [1, weight_value.shape[0]])
        y = helper.make_tensor_value_info("y", TensorProto.FLOAT, [1, weight_value.shape[1]])
        weight = numpy_helper.from_array(weight_value.astype(np.float32), name="weight")
        node = helper.make_node("MatMul", ["x", "weight"], ["y"], name="matmul")
        graph = helper.make_graph([node], "matmul_graph", [x], [y], [weight])
        if opset_imports is None:
            opset_imports = [helper.make_opsetid("", 17)]
        return helper.make_model(graph, opset_imports=opset_imports)

    def test_dynamic_weight_only_quantize_preserves_matmul_channel_order(self):
        weight_value = np.array([[0.1, 10.2], [0.3, 20.4], [0.5, 30.6]], dtype=np.float32)
        model = self._build_matmul_model(weight_value)

        quantized_model = dynamic_weight_only_quantize(model, [], [])

        dq_node = next(node for node in quantized_model.graph.node if node.op_type == "DequantizeLinear")
        initializers = {initializer.name: numpy_helper.to_array(initializer) for initializer in quantized_model.graph.initializer}
        quantized_weight = initializers[dq_node.input[0]]
        scales = initializers[dq_node.input[1]]
        zero_points = initializers[dq_node.input[2]]

        expected_scales = []
        expected_zero_points = []
        expected_quantized_columns = []
        for output_channel in weight_value.T:
            zero_point, scale, quantized_column = ort_quantize_data(
                output_channel, onnx.TensorProto.INT8, True, False
            )
            expected_scales.append(scale)
            expected_zero_points.append(zero_point)
            expected_quantized_columns.append(quantized_column)

        expected_quantized_weight = np.stack(expected_quantized_columns, axis=1)

        np.testing.assert_array_equal(scales, np.array(expected_scales, dtype=np.float32))
        np.testing.assert_array_equal(zero_points, np.array(expected_zero_points, dtype=np.int8))
        np.testing.assert_array_equal(quantized_weight, expected_quantized_weight)

    def test_dynamic_quantize_adds_missing_default_onnx_opset_import(self):
        weight_value = np.array([[0.1, 10.2], [0.3, 20.4], [0.5, 30.6]], dtype=np.float32)
        model = self._build_matmul_model(weight_value, opset_imports=[helper.make_opsetid("custom.domain", 1)])

        quantized_model = dynamic_quantize_onnx_model(model, [], [], sim_en=False)

        opset_versions = {opset.domain: opset.version for opset in quantized_model.opset_import}
        self.assertIn("", opset_versions)
        self.assertGreaterEqual(opset_versions[""], 17)
        onnx.checker.check_model(quantized_model)


if __name__ == "__main__":
    unittest.main()
